# Copyright 2023 Max Schmaltz: @maxschmaltz
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ************************************************************************


#region Imports

from dataclasses import dataclass, field
import re
from typing import List

#endregion


#region States

@dataclass(frozen=True, kw_only=True)
class State:

    '''
    ### Basic class for a Markov Chain state.

    Parameters
    ----------
    id : `int`
        number of the state in original numbering

    canonical_number : `int`
        number of the state in canonical numbering

    essential : `bool`
        whether the state is essential

    equivalency_class : `int`
        equivalency class the state belongs to

    cyclic_subclass : `int`
        cyclic subclass the state belongs to
    '''

    id:                 int
    canonical_number:   int
    essential:          bool
    equivalency_class:  int
    cyclic_subclass:    int

    def __repr__(self) -> str:
        return f'State {self.id}, canonical number: {self.canonical_number}'
    
    def __str__(self) -> str:
        return self.__repr__()


@dataclass(frozen=True, kw_only=True, repr=False)
class EssentialState(State):

    '''
    ### Class for a Markov Chain essential state.

    Parameters
    ----------
    id : `int`
        number of the state in original numbering

    canonical_number : `int`
        number of the state in canonical numbering

    equivalency_class : `int`
        equivalency class the state belongs to

    cyclic_subclass : `int`
        cyclic subclass the state belongs to
    '''

    essential: bool = field(init=False, default=True)


@dataclass(frozen=True, kw_only=True, repr=False)
class InessentialState(State):

    '''
    ### Class for a Markov Chain inessential state.

    Parameters
    ----------
    id : `int`
        number of the state in original numbering

    canonical_number : `int`
        number of the state in canonical numbering
    '''

    essential:          bool = field(init=False, default=False)
    equivalency_class:  int = field(init=False, default=None)
    cyclic_subclass:    int = field(init=False, default=None)

#endregion


#region Equivalency Classes

@dataclass(frozen=True, kw_only=True)
class CyclicSubclass:

    '''
    ### Class for a Markov Chain cyclic subclass.

    Parameters
    ----------
    id : `int`
        number of the subclass

    class_id : `int`
        equivalency class the state belongs to

    states : `list[EssentialState]`
        states that belong to the subclass
    '''

    id:         int
    class_id:   int
    states:     List[EssentialState]

    @property
    def original_states(self) -> List[int]:

        '''
        ### Returns a list of the numbers of its states in original numbering.

        Returns
        -------
        ids : `list[int]`
            list of the numbers of the states that belong to the subclass
        '''

        ids = [state.id for state in self.states]
        return ids

    @property
    def canonical_states(self) -> List[int]:

        '''
        ### Returns a list of the numbers of its states in canonical numbering.

        Returns
        -------
        ids : `list[int]`
            list of the numbers of the states that belong to the subclass
        '''

        return [state.canonical_number for state in self.states]
    
    def __len__(self) -> int:
        return len(self.states)

    def __repr__(self) -> str:
        return f'Cyclic Subclass {self.id}: states {", ".join([str(state.id) for state in self.states])}'
    
    def __str__(self) -> str:
        return self.__repr__()


@dataclass(frozen=True, kw_only=True)
class EquivalenceClass:

    '''
    ### Class for a Markov Chain equivalence class.

    Parameters
    ----------
    id : `int`
        number of the subclass

    d : `int`
        period of the class (i.e. the number of cyclic subclasses in it)

    states : `list[EssentialState]`
        states that belong to the subclass

    cyclic_subclasses : `list[CyclicSubclass]` or `None`
        list of cyclic subclasses that are in this class if `d` > 1,
        `None` otherwise
    '''

    id:                 int
    d:                  int
    states:             List[EssentialState]
    cyclic_subclasses:  List[CyclicSubclass] | None

    @property
    def original_states(self) -> List[int]:

        '''
        ### Returns a list of the numbers of its states in original numbering.

        Returns
        -------
        ids : `list[int]`
            list of the numbers of the states that belong to the class
        '''

        ids = [state.id for state in self.states]
        return ids

    @property
    def canonical_states(self) -> List[int]:

        '''
        ### Returns a list of the numbers of its states in canonical numbering.

        Returns
        -------
        ids : `list[int]`
            list of the numbers of the states that belong to the class
        '''

        return [state.canonical_number for state in self.states]

    def __len__(self) -> int:
        return len(self.states)

    def __repr__(self) -> str:

        if self.cyclic_subclasses is not None:
            prefix = 'Cyclic'
            tree = '\n' + ''.join(f'|\n|___{repr(cyclic_subclass)}\n' for cyclic_subclass in self.cyclic_subclasses)
        else: 
            prefix = 'Acyclic'
            tree = f' : states {", ".join([str(state.id) for state in self.states])}'

        return f'{prefix} Equivalency Class {self.id}{tree.rstrip()}'
    
    def __str__(self) -> str:
        return self.__repr__()
    
#endregion


@dataclass(frozen=True, kw_only=True)
class ChainStructure:

    '''
    ### Class for a Markov Chain structure.

    Parameters
    ----------
    states : `list[State]`
        states that there are in the system

    essential_states : `list[EssentialState]`
        essential states that there are in the system

    inessential_states : `list[InessentialState]`
        inessential states that there are in the system

    equivalency_classes : `list[EquivalenceClass]`
        equivalency classes of the system

    cyclic_subclasses : `list[CyclicSubclass]`
        cyclic subclasses of the system
    '''

    states:                 List[State]
    essential_states:       List[EssentialState]
    inessential_states:     List[InessentialState]
    equivalency_classes:    List[EquivalenceClass]
    cyclic_subclasses:      List[CyclicSubclass]

    @property
    def original_states(self) -> List[int]:

        '''
        ### Returns a list of the numbers of its states in original numbering.

        Returns
        -------
        ids : `list[int]`
            list of the numbers of the states that belong to the class
        '''

        ids = [state.id for state in self.states]
        return ids

    @property
    def canonical_states(self) -> List[int]:

        '''
        ### Returns a list of the numbers of its states in canonical numbering.

        Returns
        -------
        ids : `list[int]`
            list of the numbers of the states that belong to the class
        '''

        ids = [state.canonical_number for state in self.states]
        return ids
    
    @property
    def original_essential_states(self) -> List[int]:

        '''
        ### Returns a list of the numbers of its essential states in original numbering.

        Returns
        -------
        ids : `list[int]`
            list of the numbers of the essential states that belong to the class
        '''

        ids = [state.id for state in self.essential_states]
        return ids

    @property
    def canonical_essential_states(self) -> List[int]:

        '''
        ### Returns a list of the numbers of its essential states in canonical numbering.

        Returns
        -------
        ids : `list[int]`
            list of the numbers of the essential states that belong to the class
        '''

        ids = [state.canonical_number for state in self.essential_states]
        return ids
    
    @property
    def original_inessential_states(self) -> List[int]:

        '''
        ### Returns a list of the numbers of its inessential states in original numbering.

        Returns
        -------
        ids : `list[int]`
            list of the numbers of the inessential states that belong to the class
        '''

        ids = [state.id for state in self.inessential_states]
        return ids

    @property
    def canonical_inessential_states(self) -> List[int]:

        '''
        ### Returns a list of the numbers of its inessential states in canonical numbering.

        Returns
        -------
        ids : `list[int]`
            list of the numbers of the inessential states that belong to the class
        '''

        ids = [state.canonical_number for state in self.inessential_states]
        return ids

    def __len__(self) -> int:
        return len(self.states)

    def __repr__(self) -> str:
        ws = ' ' * 4
        col = re.compile('\|')
        if len(self.essential_states):
            ess_tree = f'Essential States\n{ws}' + ''.join(f'|\n{ws}|___{col.sub(f"{ws}{ws}|", repr(equivalency_class))}\n' \
                       for equivalency_class in self.equivalency_classes)
        else: ess_tree = 'Essential States: None'
        if len(self.inessential_states):
            iness_tree =  f'Inessential States: states {", ".join([str(state.id) for state in self.inessential_states])}'
        else: iness_tree = 'Inessential States: None'
        tree = f'\n|\n|___{ess_tree}|\n|___{iness_tree}'
        return f'Markov Chain{tree}'
    
    def __str__(self) -> str:
        return self.__repr__()
    

@dataclass(frozen=True, kw_only=True)
class ChainProperties:

    '''
    ### Class containing properties of a Markov Chain.

    Parameters
    ----------
    reducible : `bool`
        if the chain is reducible (i.e. has inessential states)

    polyergodic : `bool`
        if the chain is polyergodic (i.e. has several equivalency classes)

    regular : `bool` or `None`
        `True` if all the equivalency classes of the chain are acyclic;
        `False` if all the equivalency classes of the chain are cyclic;
        `None` if there are both cyclic and acyclic equivalency classes

    absorbing : `bool`
        whether all the equivalency classes of the chain are absorbing

    strong_convergence : `bool`
        whether the transition matrix converges to the end transition matrix strongly
    '''

    reducible:          bool
    polyergodic:        bool
    regular:            bool | None
    absorbing:          bool
    strong_convergence: bool

    @property
    def type_label(self) -> str:

        '''
        ### Returns label for the type of the chain based on its properties.

        Returns
        -------
        type_label : `str`
            2-place label, where "Ergodic", "Monoergodic" or "Polyergodic" stands at the first place,
            "Absorbing", "Regular", "Mixed" or "Cyclic" at the second one
        '''

        type_label = ''
        type_label += 'Ergodic' if not self.reducible else 'Monoergodic' if not self.polyergodic else 'Polyergodic'
        type_label += ' Absorbing' if self.absorbing else ' Regular' if self.regular \
                      else ' Mixed' if self.regular is None else ' Cyclic'
        return type_label

    def __repr__(self) -> str:
        props = []
        for attr, _ in self.__annotations__.items():
            value = getattr(self, attr)
            if value is None: sym = '±'
            else: sym = '+' if value else '-'
            prop = f'[{sym}{attr}]'
            props.append(prop)
        return ''.join(props)

    def __str__(self) -> str:
        return self.__repr__()


'''
### Module for computing and storing information of a discrete Markov Chain.
Markov chain is a model of a system with \(n\) states, that assumes that
transition from one state to another is independent of the history of transitions
and is strictly defined by the probability of such transition (Markov assumption).

Thus, a discrete Markov Chain is described with 2 components:

1. Initial probability distribution vector \(π = (π_0, π_1, ..., π_n)\), where \(n\) is the number 
of states in the system, and \(π_i\) is the probability that the system starts in the state \(i\).

2. Transition probability matrix \(P = (p_{ij})\), where \(p_{ij}\) is the probability to go
to the state \(j\) from the state \(i\) in one step.

Given transition probability matrix and (optional) initial state probability distribution,
`MarChie` does the following:

* computes adjacency matrix;
* computes reachability matrix (using adjacency matrix);
* computes transposed reachability matrix (using reachability);
* computes communication matrix (using reachability and transposed reachability matrices);
* computes communication matrix complement (using communication);
* computes classification matrix (using reachability and communication matrix complement);
* computes classification matrix extension (using classification matrix);
* computes equivalency classes matrix (using communication matrix and classification matrix extension);
* defines essential and inessential states;
* defines equivalency classes and their cyclic subclasses;
* builds the structure of the chain from where all its states, classes and subclasses are easily accessible;
* defines properties of the chain;
* classifies the chain (based on the properties);
* defines end behavior of the chain (based on the classification).
'''

from marchie.marchie import MarChie


author_mail = 'schmaltzmax@gmail.com'


# Copyright 2023 Max Schmaltz: @maxschmaltz
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ************************************************************************


#region Imports

import numpy as np
from typing import Tuple

#endregion


def _solve_equation_system(coefficient_mat: np.ndarray, solutions: np.ndarray) -> np.ndarray:

    '''
    ### Solves an equation system with Cramer's Rule.
    Cramer's Rule is a method to solve an equation system where
    the number of variables and the number of equations is the same.
    In the method, the matrix of the coefficients of the variables
    and the solutions of the equations are needed.

    Parameters
    ----------
    coefficient_mat : `numpy.ndarray` of shape (`n_equations`, `n_coefficients`), `n_equations` = `n_coefficients`
        matrix of the coefficients of the variables

    solutions : `numpy.ndarray` of shape (`n_equations`, )
        vector of solutions of the equations

    Returns
    -------
    pi : `numpy.ndarray` of shape (`n_equations`, )
        vector of variable values
    '''

    _, n_coefficients = coefficient_mat.shape
    pi = np.zeros((n_coefficients)) # pi is the vector of variables
    coef_delta = np.linalg.det(coefficient_mat)
    for i in range(n_coefficients):
        sub_mat = coefficient_mat.copy()
        sub_mat[:, i] = solutions.T
        delta = np.linalg.det(sub_mat)
        pi[i] = delta / coef_delta
    return pi


#region End Behavior

class EndBehavior:

    '''
    ### Base class to determine end behavior of a Markov Chain.
    End behavior of a Markov chain can be described with 3 properties:

    * end state distribution vector: what are the probabilities of
    the system to be in each state in a very large (limits to infinity) number of steps;
    accessible in `end_state_distr` property

    * end transition matrix: what are the probabilities of
    the transitions in a very large (limits to infinity) number of steps;
    accessible in `end_trans_mat` property

    * end time percentage: how much time the system will spend in each of the states
    during a very long (limits to infinity) period of time;
    accessible in `time_percentage` property

    Note that only regular chains converge to those probabilities strongly;
    cyclic and mixed chains show different empirical end behavior.

    You may refer to the demo notebook ("./demo.ipynb") to find out how the
    end behavior for different types of chains are calculated.

    Parameters
    ----------
    chain : `MarChie`
        an instantiated `MarChie` object to calculate the end behavior of
    '''

    end_state_distr: np.ndarray
    '''
    ### End state distribution vector.
    Calculates what are the probabilities of
    the system to be in each state in a very large (limits to infinity) number of steps.

    Returns
    ----------
    end_state_distr : `numpy.ndarray` of shape (`chain.n_states`, )
        end state distribution vector
    '''

    end_trans_mat: np.ndarray
    '''
    ### End transition matrix.
    Calculates what are the probabilities of
    the transitions in a very large (limits to infinity) number of steps.

    Returns
    ----------
    end_trans_mat : `numpy.ndarray` of shape (`chain.n_states`, `chain.n_states`)
        end transition matrix
    '''

    time_percentage: np.ndarray
    '''
    # End time percentage.
    Calculates how much time the system will spend in each of the states
    during a very long (limits to infinity) period of time.

    Returns
    -------
    time_percentage : `numpy.ndarray` of shape (`chain.n_states`, )
        vector of percentages (0-100) for each state in canonical numbering
    '''

    def __init__(self, chain) -> None:
        self.chain = chain
        self.end_state_distr = self._end_state_distr()
        self.end_trans_mat = self._end_trans_mat()  
        self.time_percentage = self._time_percentage()

    def equation_system(self, matrix: np.ndarray=None) -> Tuple[np.ndarray]:

        '''
        ### Converts a square matrix into an equation system.

        Parameters
        ----------
        matrix : `numpy.ndarray` of shape (`n`, `n`), optional, defaults to `None`
            the matrix to turn into an equation system; if `None`,
            convert the transition matrix of `chain`

        Returns
        -------
        coefficient_matrix : `numpy.ndarray` of shape (`n`, `n`)
            matrix of coefficients of the equation system
        
        answers : `numpy.ndarray` of shape (`n`, )
            vector of answers of the equation system

        Raises
        ------
        AssertionError
            if `matrix` is not square
        '''

        ## Below are examples for a better understanding why the system is built this way.
        ## Imagine we have 3 states, then the transition matrix looks like:
        ##   /   c00     c01     c02   \
        ##   |   c10     c11     c12   |
        ##   \   c20     c21     c22   /
        ## 
        ## If a chain fragment P (e.g. standing for an equivalent class or the whole matrix) 
        ## limits to some end matrix (fragment) Π, that end fragment Π can be represented
        ## as a matrix with equal rows π = (π0, π1, ...), in out case:
        ##   /   π0      π1      π2   \
        ##   |   π0      π1      π2   |
        ##   \   π0      π1      π2   /
        ## 
        ## That actually means that in the end behavior, the following statement will be fair:
        ## πP = π
        ##
        ## It is fair as the end fragment becomes stationary at its end.
        ## If we rewrite the equation, we get:
        ##                          /   c00     c01     c02   \
        ##   (π0      π1      π2)   |   c10     c11     c12   | = (π0      π1      π2)
        ##                          \   c20     c21     c22   /
        ##
        ## Now we can actually multiply matrices and get the equation system:
        ##   |   c00π0 + c10π1 + c20π2 = π0
        ##   |   c01π0 + c11π1 + c21π2 = π1
        ##   |   c02π0 + c12π1 + c22π2 = π2
        ##
        ## We can subtract the answers:
        ##   |   (c00 - 1)π0 + c10π1 + c20π2 = 0
        ##   |   c01π0 + (c11 - 1)π1 + c21π2 = 0
        ##   |   c02π0 + c12π1 + (c22 - 1)π2 = 0
        ##
        ## coefficient matrix looks like this then:
        ##   / c00 - 1   c10     c20   \
        ##   |   c01   c11 - 1   c21   |
        ##   \   c02     c12   c22 - 1 /
        ##
        ## As one can see, the coefficient matrix is basically just the transposed
        ## transition matrix, where the values at the diagonal are subtracted by 1
        ## (just because we subtracted 1 * πi from each equation).
        ## Now the only problem there might be is that the coefficient matrix is nondegenerate
        ## (the determiner equals 0), hence, we will not be able to solve it with Creamer's rule.
        ## To avoid that, we might substitute an equation in the system with
        ## the π vector (it's stochastic, so we know for sure that sum(π) equals 1):
        ##   |   π0    + π1    +   2         = 1
        ##   |   c01π0 + (c11 - 1)π1 + c21π2 = 0
        ##   |   c02π0 + c12π1 + (c22 - 1)π2 = 0
        ##
        ## In the end we have a coefficient matrix that equals the transposed 
        ## transition matrix where the values at the diagonal are subtracted by 1
        ## and some row is substituted by ones (since it's 1 * πi):
        ##   /    1       1        1   \
        ##   |   c01   c11 - 1   c21   |
        ##   \   c02     c12   c22 - 1 /

        if matrix is None: matrix = self.chain.trans_mat.canonical
        n_states, n = matrix.shape

        assert n_states == n, 'Only square matrices are supported.'

        coefficient_matrix = np.zeros(matrix.shape)
        coefficient_matrix = matrix.T # transpose to get coefficients
        for i in range(n_states): coefficient_matrix[i, i] -= 1 # subtract answers
        coefficient_matrix[0] = np.ones((n_states)) # substitute an equation (say, the first)
        answers = np.zeros((n_states)); answers[0] = 1 # update answers
        return coefficient_matrix, answers

    def _end_state_distr(self) -> np.ndarray:

        '''
        ### Defines end state distribution vector.
        Calculates what are the probabilities of
        the system to be in each state in a very large (limits to infinity) number of steps.

        Raises
        ------
        NotImplementedError
            if trying to instantiate the base class 
        '''
        
        raise NotImplementedError('You should define end state distribution vector in your derived class.')

    def _end_trans_mat(self) -> np.ndarray:

        '''
        ### Defines end transition matrix.
        Calculates what are the probabilities of
        the transitions in a very large (limits to infinity) number of steps.

        Raises
        ------
        NotImplementedError
            if trying to instantiate the base class 
        '''

        raise NotImplementedError('You should define end transition probability matrix in your derived class.')

    def _time_percentage(self) -> np.ndarray:

        '''
        ### Defines end time percentage.
        Calculates how much time the system will spend in each of the states
        during a very long (limits to infinity) period of time.

        Returns
        -------
        time_percentage : `numpy.ndarray` of shape (`chain.n_states`, )
            vector of percentages (0-100) for each state in canonical numbering
        '''
        
        time_percentage = self.end_state_distr * 100
        return time_percentage
    

class ErgodicChainEndBehavior(EndBehavior):

    '''
    ### Class to determine end behavior of an ergodic (irreducible) Markov Chain.
    An ergodic (irreducible) Markov Chain has no inessential states and 
    a single equivalency class (and thus coincide with it).

    Refer to the base class `EndBehavior` for parameters and description 
    of what the end behavior is.
    '''

    def _end_state_distr(self) -> np.ndarray:

        '''
        ### Defines end state distribution vector.
        Calculates what are the probabilities of
        the system to be in each state in a very large (limits to infinity) number of steps.

        Returns
        ----------
        end_state_distr : `numpy.ndarray` of shape (`chain.n_states`, )
            end state distribution vector
        '''

        end_state_distr = _solve_equation_system(
            *self.equation_system()
        )
        return end_state_distr

    def _end_trans_mat(self) -> np.ndarray:

        '''
        ### Defines end transition matrix.
        Calculates what are the probabilities of
        the transitions in a very large (limits to infinity) number of steps.

        Returns
        ----------
        end_trans_mat : `numpy.ndarray` of shape (`chain.n_states`, `chain.n_states`)
            end transition matrix
        '''

        end_trans_mat = self.chain._zero_mask
        end_trans_mat[:] = self.end_state_distr
        return end_trans_mat


class MonoergodicChainEndBehavior(EndBehavior):

    '''
    ### Class to determine end behavior of a monoergodic Markov Chain.
    A monoergodic Markov Chain has \(1\) or more inessential states and 
    a single equivalency class.

    Refer to the base class `EndBehavior` for parameters and description 
    of what the end behavior is.
    '''

    def _end_state_distr(self) -> np.ndarray:

        '''
        ### Defines end state distribution vector.
        Calculates what are the probabilities of
        the system to be in each state in a very large (limits to infinity) number of steps.

        Returns
        ----------
        end_state_distr : `numpy.ndarray` of shape (`chain.n_states`, )
            end state distribution vector
        '''

        essential_submat = self.chain.trans_mat.canonical[
            # left upper corner given the canonical numbering
            0: self.chain.n_essential_states,
            0: self.chain.n_essential_states
        ]
        essential_end_state_distr = _solve_equation_system(
            *self.equation_system(essential_submat)
        )
        end_zeros = np.zeros((self.chain.n_inessential_states))
        end_state_distr = np.concatenate((essential_end_state_distr, end_zeros), axis=0)
        return end_state_distr

    def _end_trans_mat(self) -> np.ndarray:

        '''
        ### Defines end transition matrix.
        Calculates what are the probabilities of
        the transitions in a very large (limits to infinity) number of steps.

        Returns
        ----------
        end_trans_mat : `numpy.ndarray` of shape (`chain.n_states`, `chain.n_states`)
            end transition matrix
        '''

        end_trans_mat = self.chain._zero_mask
        end_trans_mat[:] = self.end_state_distr
        return end_trans_mat


class PolyergodicChainEndBehavior(EndBehavior):

    '''
    ### Class to determine end behavior of a polyergodic Markov Chain.
    A polyergodic Markov Chain has \(1\) or more inessential states and 
    several (\(2\) or more) equivalency classes.

    Refer to the base class `EndBehavior` for parameters and description 
    of what the end behavior is.
    '''
  
    def __init__(self, chain) -> None:
        self.chain = chain
        self.end_trans_mat = self._end_trans_mat()  
        self.end_state_distr = self._end_state_distr()
        self.time_percentage = None

    def _end_state_distr(self) -> np.ndarray:

        '''
        ### Defines end state distribution vector.
        Calculates what are the probabilities of
        the system to be in each state in a very large (limits to infinity) number of steps.

        Returns
        ----------
        end_state_distr : `numpy.ndarray` of shape (`chain.n_states`, )
            end state distribution vector
        '''

        end_state_distr = np.matmul(
            self.chain.init_distr,
            self.end_trans_mat
        )
        return end_state_distr

    def _end_trans_mat(self) -> np.ndarray:

        '''
        ### Defines end transition matrix.
        Calculates what are the probabilities of
        the transitions in a very large (limits to infinity) number of steps.

        Returns
        ----------
        end_trans_mat : `numpy.ndarray` of shape (`chain.n_states`, `chain.n_states`)
            end transition matrix
        '''

        c = 0

        # essential submatrix `Пe`
        essential_submat = np.zeros((self.chain.n_essential_states, self.chain.n_states))
        essential_vectors = []
        for i, eq_class in enumerate(self.chain.structure.equivalency_classes):

            # transitions in each class `t` are described by its submatrix `Tt`
            begin, end = eq_class.canonical_states[0], eq_class.canonical_states[-1]
            class_mat = self.chain.trans_mat.canonical[
                begin: end + 1,
                begin: end + 1
            ]

            # each row is a vector of `π` form `(0, ..., π_, ..., 0)`,
            # `π_` is a subvector such that `π_ x Tt = π_`;
            # here we find `π_`
            class_vec = _solve_equation_system(
                *self.equation_system(class_mat)
            )
            # number of zeros on the left from `π_`
            n_zeros_left = sum([
                len(prev_eq_class) 
                for prev_eq_class in self.chain.structure.equivalency_classes[:i]
            ])
            # number of zeros on the right from `π_`
            n_zeros_right = sum([
                len(next_eq_class) 
                for next_eq_class in self.chain.structure.equivalency_classes[i + 1:]
            ]) + self.chain.n_inessential_states

            # finally `π`
            class_vec = np.concatenate((
                np.zeros((n_zeros_left)),
                class_vec,
                np.zeros((n_zeros_right))
            ), axis=0)

            # put in `Пe`; store `π`
            for _ in range(len(eq_class)): essential_submat[c] = class_vec; c += 1;
            essential_vectors.append(class_vec)


        # inessential to essential submatrix `B`
        absorbing_trans_mat = self.chain.trans_mat.canonical.copy()
        states_to_delete = []
        for eq_class in self.chain.structure.equivalency_classes:
            # leave the first state of each class
            states_to_delete += eq_class.canonical_states[1:]
            # substitute the probabilities with a single sum of them
            absorbing_trans_mat[:, eq_class.canonical_states[0]] = (absorbing_trans_mat[:, eq_class.canonical_states]).sum(axis=1)
        # delete remaining states
        absorbing_trans_mat = np.delete(absorbing_trans_mat, states_to_delete, axis=0)
        absorbing_trans_mat = np.delete(absorbing_trans_mat, states_to_delete, axis=1)

        absorbing_n_states, _ = absorbing_trans_mat.shape
        # each equivalency class we substituted with a single state,
        # which means in `absorbing_trans_mat` the number of essential states
        # equals the number of equivalency classes in the original matrix
        absorbing_identity_mat = np.identity(self.chain.n_inessential_states)
        absorbing_inessential_submat = absorbing_trans_mat[
            # right lower corner given the canonical numbering
            self.chain.n_equivalency_classes: absorbing_n_states,
            self.chain.n_equivalency_classes: absorbing_n_states
        ]
        absorbing_end_inessential_submat = np.linalg.inv(absorbing_identity_mat - absorbing_inessential_submat)
        absorbing_inessential_to_essential_submat = absorbing_trans_mat[
            # left lower corner given the canonical numbering
            self.chain.n_equivalency_classes: absorbing_n_states,
            0: self.chain.n_equivalency_classes
        ]
        end_inessential_to_essential_submat = np.matmul(
            absorbing_end_inessential_submat,
            absorbing_inessential_to_essential_submat
        )


        # inessential submatrix `Пie`
        inessential_submat = np.zeros((self.chain.n_inessential_states, self.chain.n_states))
        for i, iness_row in enumerate(inessential_submat):
            for t, ess_row in enumerate(essential_vectors):
                iness_row += end_inessential_to_essential_submat[i, t] * ess_row


        end_trans_mat = np.concatenate((essential_submat, inessential_submat), axis=0)
        return end_trans_mat

    def _time_percentage(self) -> np.ndarray:

        '''
        ### Defines end time percentage.
        Calculates how much time the system will spend in each of the states
        during a very long (limits to infinity) period of time.
        NB! Not supported for polyergodic chains.

        Raises
        -------
        NotImplementedError
            when trying to calculate end time percentage for a polyergodic Markov Chain
        '''

        raise NotImplementedError(
            'Percentages of time for each state are not supposed to be calculated for polyergodic chains.'
        )
  

class PolyergodicAbsorbingChainEndBehavior(PolyergodicChainEndBehavior):

    '''
    ### Class to determine end behavior of a polyergodic absorbing Markov Chain.
    A polyergodic absorbing Markov Chain has 1 or more inessential states and 
    several (2 or more) equivalency classes, each of which is absorbing.

    Refer to the base class `EndBehavior` for parameters and description 
    of what the end behavior is.
    '''

    def _end_trans_mat(self) -> np.ndarray:

        '''
        ### Defines end transition matrix.
        Calculates what are the probabilities of
        the transitions in a very large (limits to infinity) number of steps.

        Returns
        ----------
        end_trans_mat : `numpy.ndarray` of shape (`chain.n_states`, `chain.n_states`)
            end transition matrix
        '''

        # essential submatrix `T`
        end_essential_submat = np.identity(self.chain.n_essential_states)

        # inessential to essential submatrix `B`
        identity_mat = np.identity(self.chain.n_inessential_states)
        inessential_submat = self.chain.trans_mat.canonical[
            # right lower corner given the canonical numbering
            self.chain.n_essential_states: self.chain.n_states,
            self.chain.n_essential_states: self.chain.n_states
        ]
        end_inessential_submat = np.linalg.inv(identity_mat - inessential_submat)
        inessential_to_essential_submat = self.chain.trans_mat.canonical[
            # left lower corner given the canonical numbering
            self.chain.n_essential_states: self.chain.n_states,
            0: self.chain.n_essential_states
        ]
        end_inessential_to_essential_submat = np.matmul(end_inessential_submat, inessential_to_essential_submat)

        end_trans_mat = self.chain._zero_mask
        end_trans_mat[
            # left upper corner given the canonical numbering
            0: self.chain.n_essential_states,
            0: self.chain.n_essential_states
        ] = end_essential_submat
        end_trans_mat[
            # left lower corner given the canonical numbering
            self.chain.n_essential_states: self.chain.n_states,
            0: self.chain.n_essential_states
        ] = end_inessential_to_essential_submat

        return end_trans_mat
    
#endregion


# Copyright 2023 Max Schmaltz: @maxschmaltz
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ************************************************************************


#region Imports

import numpy as np
import graphviz
from typing import Tuple, Dict

from marchie.structure import *
from marchie.end_behavior import *

#endregion


class MarChie:
    
    '''
    ### Class for computing and storing information of a discrete Markov Chain.
    Markov chain is a model of a system with \(n\) states, that assumes that
    transition from one state to another is independent of the history of transitions
    and is strictly defined by the probability of such transition (Markov assumption).

    Thus, a discrete Markov Chain is described with 2 components:

    1. Initial probability distribution vector \(\pi = (\pi_0, \pi_1, ..., \pi_n)\), where \(n\) is the number 
    of states in the system, and \(\pi_i\) is the probability that the system starts in the state \(i\).

    2. Transition probability matrix \(P = (p_{ij})\) where \(p_{ij}\) is the probability to go
    to the state \(j\) from the state \(i\) in one step.

    Given transition probability matrix and (optional) initial state probability distribution,
    `MarChie` does the following:

    * computes adjacency matrix;
    * computes reachability matrix (using adjacency matrix);
    * computes transposed reachability matrix (using reachability);
    * computes communication matrix (using reachability and transposed reachability matrices);
    * computes communication matrix complement (using communication);
    * computes classification matrix (using reachability and communication matrix complement);
    * computes classification matrix extension (using classification matrix);
    * computes equivalency classes matrix (using communication matrix and classification matrix extension);
    * defines essential and inessential states;
    * defines equivalency classes and their cyclic subclasses;
    * builds the structure of the chain from where all its states, classes and subclasses are easily accessible;
    * defines properties of the chain;
    * classifies the chain (based on the properties);
    * defines end behavior of the chain (based on the classification).

    Parameters
    ----------
    init_distr : `numpy.ndarray` of shape (`n_states`, ), optional, defaults to `None`
        initial probability distribution vector; if not provided, generated automatically

    trans_mat : `numpy.ndarray` of shape (`n_states`, `n_states`)
        transition probabilities matrix

    Raises
    ------
    AssertionError
        if chain components are not valid

    Notes
    -----
    The vector of initial state distribution should add up to \(1\) as the system 
    must begin at some state anyway, so the sum of the probabilities of all the states must be \(1\);
    otherwise the vector is invalid.

    The rows of the transition matrix should add up to \(1\) as from each step, the system must be able to go
    to some state anyway, so the sum of the probabilities of the transitions will be \(1\);
    otherwise the matrix is invalid.
    '''

    class _matrix(np.ndarray):

        '''
        ### Wrapper class for Markov Chain matrices.
        Inherits from `numpy.ndarray`; the main feature is implementing
        `canonical` property that rebuilds the input matrix
        in accordance with canonical numbering.

        Parameters
        ----------
        input_matrix : `numpy.ndarray` of shape `input_shape`
            input matrix to wrap around

        canonical_mapping: `Dict[int, int]`
            mapping between state and its canonical number

        Returns
        -------
        obj : `_matrix` of shape `input_shape`
            wrapped matrix
        
        Notes
        -----
        In general, class adds 3 properties:

        canonical_mapping: `Dict[int, int]`
            mapping between state and its canonical number
        
        dim: `int`
            dimentionality of matrix

        canonical: `numpy.ndarray` of shape `input_shape`
            input matrix where the states renumbered canonically
        '''

        def __new__(cls, input_matrix: np.ndarray, canonical_mapping: Dict[int, int]) -> np.ndarray:
            obj = np.asarray(input_matrix).view(cls)
            obj.canonical_mapping = canonical_mapping
            obj.original = input_matrix
            obj.dim = len(obj.shape)
            return obj

        @property
        def canonical(self) -> np.ndarray:

            '''
            ### Rebuilds the input matrix in accordance with canonical numbering.

            Returns
            -------
            can_matrix: `numpy.ndarray` of shape `input_shape`
                input matrix where the states renumbered canonically
            '''

            can_matrix = self.copy()
            for i, row in enumerate(self):
                if self.dim == 1: # for state distribution
                    i_ = self.canonical_mapping[i]
                    can_matrix[i_] = row
                else: # for 2D matrices
                    for j, value in enumerate(row):
                        i_ = self.canonical_mapping.get(i)
                        j_ = self.canonical_mapping.get(j)
                        if i_ is None or j_ is None: continue # for extension
                        can_matrix[i_, j_] = value
            if self.dim == 1: can_matrix = can_matrix.squeeze()
            return can_matrix

    init_distr: _matrix
    '''
    ### Initial state distribution vector.
    Vector \(\pi = (\pi_0, \pi_1, ..., \pi_n)\), where \(n\) is the number 
    of states in the system, and \(\pi_i\) is the probability that the system starts in the state \(i\).

    Returns
    -------
    init_distr : `March._matrix` of shape (`n_states`, )
        distribution vector
    '''

    trans_mat: _matrix
    '''
    ### Transition probability matrix
    Matrix \(P = (p_{ij})\) where \(p_{ij}\) is the probability to go
    to the state \(j\) from the state \(i\) in one step.

    Returns
    -------
    trans_mat : `March._matrix` of shape (`n_states`, `n_states`)
        transition matrix
    '''

    adjacency_mat: _matrix
    '''
    ### Adjacency matrix of the chain.
    Adjacency matrix is a matrix that indicates existence of one-step paths between states,
    where a one-step path from the state \(i\) to the state \(j\) exists 
    if there a non-zero probability transition between them.

    Returns
    -------
    adjacency_mat : `March._matrix` of shape (`n_states`, `n_states`)
        boolean-like matrix indicating one-step paths between states,
        where `adjacency_mat[i, j]` is \(1\) if the probability of tranition 
        from the state `i` to the state `j` is not zero, else \(0\)

    Notes
    -----
    Adjacency matrix is calculated as shown below:

    $$
    A = (a_{ij}) | a_{ij} =
    \\begin{cases}
        1, & p_{ij} > 0     \\\\
        0, & p_{ij} = 0
    \\end{cases}
    $$

    where \(A\) is adjacency matrix, \((p_{ij})\) is transition matrix.
    '''

    reachability_mat: _matrix
    '''
    ### Reachability matrix of the chain.
    Reachability matrix is a matrix that indicates existence of paths between states,
    where a path from the state \(i\) to the state \(j\) exists if there a sequence 
    of non-zero probability transitions between states such that it leads from \(i\) to \(j\)
    (e.g. \(i\) → \(h\) → \(k\) → ... → \(j\)).

    Returns
    -------
    reachability_mat : `March._matrix` of shape (`n_states`, `n_states`)
        boolean-like matrix indicating paths between states,
        where `reachability_mat[i, j]` is \(1\) if there is a path 
        from the state `i` to the state `j`, else \(0\)

    Notes
    -----
    Reachability matrix is calculated as shown below:

    $$
    D = (d_{ij}) | d_{ij} =
    \\begin{cases}
        1, & \\exists i → j        \\\\
        0, & \\neg \exists i → j
    \\end{cases}
    $$

    where \(D\) is reachability matrix, \(→\) marks a path.
    '''

    reachability_mat_tr: _matrix
    '''
    ### Transposed reachability matrix of the chain.
    Reachability matrix is a matrix that indicates existence of inverse paths between states,
    where a path from the state \(i\) to the state \(j\) exists if there a sequence 
    of non-zero probability transitions between states such that it leads from \(i\) to \(j\)
    (e.g. \(i\) → \(h\) → \(k\) → ... → \(j\)).

    Returns
    -------
    reachability_mat_tr : `March._matrix` of shape (`n_states`, `n_states`)
        boolean-like matrix indicating paths between states,
        where `reachability_mat_tr[i, j]` is \(1\) if there is a path 
        from the state `j` to the state `i`, else \(0\)

    Notes
    -----
    Transposed reachability matrix is calculated as shown below:

    $$
    D^T = (d^T_{ij}) | d^T_{ij} =
    \\begin{cases}
        1, & \\exists j → i        \\\\
        0, & \\neg \exists j → i
    \\end{cases}
    $$

    where \(D^T\) is transposed reachability matrix, \(→\) marks a path.
    '''

    communication_mat: _matrix
    '''
    ### Communication matrix of the chain.
    Communication matrix is a matrix that indicates existence of bidirectional paths between states,
    where a bidirectional path between the states \(i\) and \(j\) exists if there a sequence 
    of non-zero probability transitions from \(i\) to \(j\) (e.g. \(i\) → \(h\) → \(k\) → ... → \(j\))
    and from \(j\) to \(i\) (not exactly the same way back).

    Returns
    -------
    communication_mat : `March._matrix` of shape (`n_states`, `n_states`)
        boolean-like matrix indicating bidirectional paths between states,
        where `communication_mat[i, j]` is \(1\) if there is a path 
        from the state `j` to the state `i` and from `j` to `i`, else \(0\)

    Notes
    -----
    Communication matrix is calculated as shown below:

    $$
    C = (c_{ij}) = D \\times D^T | c_{ij} = 
    \\begin{cases}
        1, & d_{ij} = d_{ji} = 1        \\\\
        0, & \\neg (d_{ij} = d_{ji} = 1)
    \\end{cases}
    $$

    where \(C\) is communication matrix, \(D\) is reachability matrix,
    \(D^T\) is transposed reachability matrix, \(→\) marks a path.
    '''

    communication_mat_comp: _matrix
    '''
    ### Communication matrix complement. 
    Communication matrix complement is a matrix that indicates nonexistence of bidirectional paths between states,
    where a bidirectional path between the states \(i\) and \(j\) exists if there a sequence 
    of non-zero probability transitions from \(i\) to \(j\) (e.g. \(i\) → \(h\) → \(k\) → ... → \(j\))
    and from \(j\) to \(i\) (not exactly the same way back).

    Returns
    -------
    communication_mat_comp : `March._matrix` of shape (`n_states`, `n_states`)
        boolean-like matrix indicating absence of bidirectional paths between states,
        where `communication_mat[i, j]` is \(1\) if there is no path 
        from the state `j` to the state `i` and from `j` to `i`, else 0

    Notes
    -----
    Communication matrix complement is calculated as follows:

    $$
    \\overline C = (\\overline c_{ij}) | \\overline c_{ij} = 
    \\begin{cases}
        1, & c_{ij} = 0     \\\\
        0, & c_{ij} = 1
    \\end{cases}
    $$

    where \(\\overline C\) is communication matrix complement, \(c_{ij}\) is communication matrix.
    '''

    classification_mat: _matrix
    '''
    ### Classification matrix of the chain.
    Classification matrix is a step for building extended classification matrix.

    Returns
    -------
    classification_mat : `March._matrix` of shape (`n_states`, `n_states`)
        boolean-like matrix where `classification_mat[i, j]` is \(1\) if there is a path 
        from the state `i` to the state `j` but states `j` and `i` do not communicate, 0 otherwise

    Notes
    -----
    Classification matrix is calculated as follows:

    $$
    T = D \\times \\overline C = (t_{ij}) | t_{ij} = 
    \\begin{cases}
        1, & d_{ij} = \\overline c_{ij} = 1      \\\\
        0, & \\neg (d_{ij} = \\overline c_{ij} = 1)
    \\end{cases}
    $$

    where \(T\) is classification matrix, \(d_{ij}\) is reachability matrix, 
    \((\\overline c_{ij})\) is communication matrix complement.
    '''

    classification_mat_ext: _matrix
    '''
    ### Classification matrix extension. 
    Classification matrix extension indicates essential and inessential states.
    Essential states are the ones that communicate with each (!) state they have a path to;
    respectively, the ineccential are the ones for which this statement is not true.

    Returns
    -------
    classification_mat_ext : `March._matrix` of shape (`n_states`, `n_states` + 1)
        boolean-like matrix where the last column indicates essentiality of states:
        if classification_mat_ext[`i`, `n_states` + 1] equals 1, the state `i`
        is inessential, if 0, is essential

    Notes
    -----
    Classification matrix extention is defined below:

    $$
    T_{ext} = T|(t_{i(N + 1)}) | t_{i(N + 1)} = 
    \\begin{cases}
        1, & \\exists j t_{ij} = 1        \\\\
        0, & \\neg \\exists j t_{ij} = 1
    \\end{cases}
    $$

    where \(T_{ext}\) is classification matrix extension, \(T\) is classification matrix,
    \(N\) is the number of states in the system.
    '''

    equivalency_cls_mat: _matrix
    '''
    ### Equivalency matrix of the chain.

    Returns
    -------
    equivalency_cls_mat : `March._matrix` of shape (`n_classes`, `n_essential`)
        matrix showing belonging of each essential (!) state to its equivalence class;
        `equivalency_cls_mat[i, j]` indicates that the state `j` is in `i`th equivalence class

    Notes
    -----
    Matrix of equivalency classes is defined as follows:

    $$
    K = (k_{ij}) | k_{it} \\Longleftrightarrow i \\in EC_t
    $$

    where \(K\) is equivalency matrix.
    '''

    structure: ChainStructure
    '''
    ### Structure of the chain.
    That includes:

    * Essential and inessential states:
    Essential state is such a state \(i\) that communicates with every (!) state \(j\)
    it has a path to; respectively, a state is inessential if that is not true.

    * Cyclic subclasses of equivalence classes:
    Each equivalence classes has a period \(d\): it equals the GCD of the lengths
    of all the contours going through a state \(i\) belonging to the equivalence class.
    Contour is a cyclic path from a state to itself (e.g. \(i\) → \(h\) → \(k\) → ... → \(i\)), 
    its length is the number of the transitions within it.
    It is proven that all the states of an equivalence class have the same period \(d\).
    If \(d = 1\), the equivalence class is acyclic.
    If \(d \geq 2\), it can be split into \(d\) cyclic subclasses. Cyclic subclass \(C_r\) of equivalence class \(EC_n\)
    is a subset of states of \(EC_n\) such that 
    
    * the states are not adjacent (i.e. there is no one-step path between them);
    * from any of the states there is a one-step path only to a state of the next cyclic subclass \(C_{r + 1}\);
    * the states belong to an only cyclic subclass;
    * the equivalence class is fully distributed between cyclic subclasses (no states left outside of the subclasses).

    That means the cyclic subclasses are closed and form a cycle themselves:

    \(C_0 → C_1 → C_2 → ... → C_{d - 1} → C_0\)

    * Canonical numbers of the states in the system structure:
    Canonical numbering is a numbering that goes in strict order through each state in each cyclic subclass
    of each equivalency class, followed only then by inessential states. That allows to transform the transition matrix
    into a block matrix that represents probabilities of transition between cyclic subclasses.

    Returns
    -------
    structure : `ChainStructure`
        class containing attributes for chain structure

    Notes
    -----
    Essentiality of states if defined as follows:
    $$
    i \\in S_e \\iff 
    \\forall j: \\quad i → j \\Longrightarrow j → i
    $$
    $$
    i \\in S_{ie} \\iff 
    \\neg \\forall j: \\quad i → j \\Longrightarrow j → i
    $$

    Where \(S_e\) is the set of essential states, \(S_{ie}\) is the set of inessential states.

    Formally, the period of an equivalency class is found as:

    $$
    d = \\text{GCD}(|(i → i)_1|, ..., |(i → i)_r|)
    $$
        
    where \((i → i)\) is a contour, \(|(i → i)_1|\) is a length of a contour

    If \(d\) equials 1, the equivalence class is acyclic, which means it has no cyclic subclasses.

    Canonical numbering goes as follows:

    \(EC_1\):

    \(C_0: \\quad 1, \\quad 2, \\quad 3, \\quad ..., \\quad r_0\)

    \(C_1: \\quad r_0 + 1, \\quad r_0 + 2, \\quad ..., \\quad r_1\)

    \(...\)

    \(C_{d - 1}: r_0 + r_1 + ... + r_{d - 1} + 1, \\quad r_0 + r_1 + ... + r_{d - 1} + 2, 
    \\quad ..., \\quad r_0 + r_1 + ... + r_{d - 1} + r_{d} = N_{EC_1}\)

    \(EC_2\): (same way)

    \(...\)

    \(EC_L\): (same way)

    \(S_{ie}: \\quad N_e + 1, \\quad N_e + 2, \\quad ..., \\quad N_e + N_{ie} = N\)

    where \(EC_k\) is equivalence class \(k\), \(C_s\) is a cyclic subclass \(s\) of the equivalence class,
    \(N_{EC_k}\) is number of states in equivalence class \(k\), \(N_e\) is number of essential states in the system,
    \(N\) is number of states in the system.
    '''

    properties: ChainProperties
    '''
    Markov Chain classification:

    1. \(S\) = \(Se\) = \(EC\) ⇒ ergodic / irreducible Markov Chain

        1.1. \(EC\) is regular ⇒ regular irreducible Markov Chain

        \(\quad\) 1.1.1. \(EC\) is absorbing ⇒ absorbing irreducible Markov Chain

        1.2. \(EC\) is cyclic ⇒ cyclic irreducible Markov Chain

    2. \(Sie ≠ ∅\) ⇒ reducible Markov Chain

        2.1. a single \(EC\) ⇒ monoergodic Markov Chain

        \(\quad\)2.1.1. the single \(EC\) is regular ⇒ regular monoergodic Markov Chain

        \(\qquad\) 2.1.1.1. the single \(EC\) is absorbing ⇒ absorbing monoergodic Markov Chain

        \(\quad\) 2.1.2. the single \(EC\) is cyclic ⇒ cyclic monoergodic Markov Chain

        2.2. more than one \(EC\)s ⇒ polyergodic Markov Chain

        \(\quad\) 2.2.1. all the \(EC\)s are regular ⇒ regular polyergodic Markov Chain

        \(\qquad\) 2.2.1.1. all the \(EC\)s are absorbing ⇒ absorbing polyergodic Markov Chain

        \(\quad\) 2.2.2. all the \(EC\)s are cyclic ⇒ cyclic polyergodic Markov Chain

        \(\quad\) 2.2.3. there are both regular and cyclic \(EC\)s ⇒ mixed polyergodic Markov Chain,

    where \(S\) is set of states, \(Se\) is set of essential states, \(Sie\) is set of inessential states,
    \(EC\) stands for equivalence class.

    Returns
    -------
    structure : `ChainProperties`
        class containing attributes for chain properties

    Notes
    -----
    Note that in our classification we won't just return a single label of the class, but rather
    use a more "linguistic" notation: we will define a set of properties, each of which can be 
    either `True`, or `False`, or `None` (if not applicable). Thus, we will be able to 
    split the space of classes uniquely, yet be able to access chain features in a more convenient way.

    Reminder: a \(EC\) is regular if it has a period \(d\) = 1, i.e. has no cyclic subclasses, otherwise cyclic.
    Reminder: a \(EC\) is absorbing if it contains a single state.
    
    The properties are:

    * `reducible` (if there are inessential states)
    * `polyergodic` (if there are more than one `EC`)
    * `regular` (cyclic if `False`, mixed if `None`)
    * `absorbing` (may be `True` only if regular)
    * `strong_convergence` (if the matrix strongly converges to a final matrix in a great number of steps)

    The matrix converges strongly if it's regular, weakly otherwise.

    Thus, we'll have a "linguistic-like" notation, e.g. a cyclic polyergodic Markov Chain
    could be notated as ```[+reducible][+polyergodic][-regular][∅absorbing][-strong_convergence]```.
    '''

    end_behavior: EndBehavior
    '''
    End behavior of a Markov chain can be described with 3 properties:

    * end state distribution vector: what are the probabilities of
    the system to be in each state in a very large (limits to infinity) number of steps;
    accessible in `end_state_distr` property

    * end transition matrix: what are the probabilities of
    the transitions in a very large (limits to infinity) number of steps;
    accessible in `end_trans_mat` property

    * end time percentage: how much time the system will spend in each of the states
    during a very long (limits to infinity) period of time;
    accessible in `time_percentage` property

    Returns
    -------
    end_behavior : `EndBehavior`
        class containing attributes for end behavior components

    Notes
    -----
    that only regular chains converge to those probabilities strongly;
    cyclic and mixed chains show different empirical end behavior.
    '''

    def __init__(self, *, init_distr: np.ndarray=None, trans_mat: np.ndarray) -> None:

        if init_distr is None: init_distr, _ = self.generate_chain(n_states=trans_mat.shape[0])
        
        self._check_chain(init_distr, trans_mat) # check if the components are valid

        self.init_distr = init_distr # initial states distribution
        self.trans_mat = trans_mat # transition matrix, i.d. P(1)
        self.adjacency_mat = self._adjacency_mat() # adjacency matrix, i.e. A
        self.reachability_mat = self._reachability_mat() # reachability matrix, i.e. D
        self.reachability_mat_tr = self._reachability_mat_tr() # transposed reachability matrix, i.e. DT
        self.communication_mat = self._communication_mat() # communication matrix, i.e. C
        self.communication_mat_comp = self._communication_mat_comp() # communication matrix complement, i.e. Ccomp
        self.classification_mat = self._classification_mat() # classification matrix, i.e. T
        self.classification_mat_ext = self._classification_mat_ext() # classification matrix extension, i.e. Text
        
        self.equivalency_cls_mat = self._build_structure() # builds `structure``, equivalency classes matrix

        # canonize matrices: add them `canonical` property
        cm = {'canonical_mapping': self.canonical_mapping}
        def canonize(matrix): return self._matrix(input_matrix=matrix, **cm)
        self.init_distr = canonize(self.init_distr) # initial states distribution
        self.trans_mat = canonize(self.trans_mat) # transition matrix, i.d. P(1)
        self.adjacency_mat = canonize(self.adjacency_mat) # adjacency matrix, i.e. A
        self.reachability_mat = canonize(self.reachability_mat) # reachability matrix, i.e. D
        self.reachability_mat_tr = canonize(self.reachability_mat_tr) # transposed reachability matrix, i.e. DT
        self.communication_mat = canonize(self.communication_mat) # communication matrix, i.e. C
        self.communication_mat_comp = canonize(self.communication_mat_comp) # communication matrix complement, i.e. Ccomp
        self.classification_mat = canonize(self.classification_mat) # classification matrix, i.e. T
        self.classification_mat_ext = canonize(self.classification_mat_ext) # classification matrix extension, i.e. Text
        self.equivalency_cls_mat = canonize(self.equivalency_cls_mat) # equivalency classes matrix

        self._classify() # builds `properties` property

        if not self.properties.reducible: end_behavior = ErgodicChainEndBehavior
        elif not self.properties.polyergodic: end_behavior = MonoergodicChainEndBehavior
        elif not self.properties.absorbing: end_behavior = PolyergodicChainEndBehavior
        else: end_behavior = PolyergodicAbsorbingChainEndBehavior

        self.end_behavior = end_behavior(self) # define end behavior

    #region Properties

    @property
    def n_states(self) -> int:

        '''
        ### Returns the number of states of the system.

        Returns
        -------
        n_states : `int`
            number of states
        '''

        n_states, n_states = self.trans_mat.shape
        return n_states

    @property
    def n_essential_states(self) -> int:

        '''
        ### Returns the number of essential states of the system.

        Returns
        -------
        n_essential_states : `int`
            number of essential states
        '''

        n_essential_states = len(self.structure.essential_states)
        return n_essential_states
    
    @property
    def n_inessential_states(self) -> int:

        '''
        ### Returns the number of inessential states of the system.

        Returns
        -------
        n_inessential_states : `int`
            number of inessential states
        '''

        n_inessential_states = len(self.structure.inessential_states)
        return n_inessential_states
    
    @property
    def n_equivalency_classes(self) -> int:

        '''
        ### Returns the number of equivalency classes of the system.

        Returns
        -------
        n_eq_classes : `int`
            number of equivalency classes
        '''

        n_eq_classes = len(self.structure.equivalency_classes)
        return n_eq_classes
    
    @property
    def n_cyclic_subclasses(self) -> int:

        '''
        ### Returns the number of all cyclic subclasses (from all equivalency classes) of the system.

        Returns
        -------
        n_subclasses : `int`
            number of equivalency classes
        '''

        n_subclasses = len(self.structure.cyclic_subclasses)
        return n_subclasses

    @property
    def shape(self) -> Tuple[int, int]:

        '''
        ### Returns the shape of the transition matrix.

        Returns
        -------
        shape : `tuple[int, int]`
            shape of the transition matrix
        '''

        shape = (self.n_states, self.n_states)
        return shape
    
    @property
    def canonical_mapping(self) -> Dict[int, int]:

        '''
        ### Returns mapping between original states and canonical numbers.

        Returns
        -------
        mapping : `dict[int, int]`
            mapping between original states and canonical numbers
        '''
        
        # return {
        #     state.id: state.canonical_number
        #     for state in self.structure.states
        # }
        mapping = dict(zip(
            self.structure.original_states,
            self.structure.canonical_states
        ))
        return mapping

    #endregion

    #region Utils

    @classmethod
    def generate_chain(cls, n_states: int) -> Tuple[np.ndarray]:

        '''
        ### Generates a transition matrix and an initial distribution.
        
        Parameters
        ----------
        n_states : `dict`
            number of states in the chain 

        Returns
        -------
        init_distr : `numpy.ndarray` of shape (`n_states`, )
            vector of initial distribution of probabilities of the states

        trans_mat : `numpy.ndarray` of shape (`n_states`, `n_states`)
            transition matrix, i.e. matrix of probabilities of transitions between states,
            where `trans_mat[i, j]` is probability of transition from the `i`th state to the `j`th state in a single step
        '''

        init_distr = np.random.rand(n_states)
        init_distr /= init_distr.sum()

        trans_mat = np.random.rand(n_states, n_states)
        for row in trans_mat: row /= row.sum()

        return init_distr, trans_mat

    def _check_chain(self, init_distr: np.ndarray, trans_mat: np.ndarray) -> None:

        '''
        ### Checks if matrices are valid components of a Markov chain.
        A Markov chain should have 2 components a transition matrix i.e. matrix of probabilities of transitions 
        between states and an initial distribution of probabilities of the states. Since the transition matrix 
        is a matrix of probabilities, the sum of the probabilities in each row should add up to \(1\); 
        the initial distribution probabilities should do so altogether.

        Raises
        ------
        AssertionError
            if any of the probabilities is out of the range \([0; 1]\).

        AssertionError
            if the shape of the initial distribution vector is not `(n, )`
            or the shape of the transition matrix is not `(n, n)`

        AssertionError
            if the probabilities of the initial distribution do not add up to \(1\)

        AssertionError
            if the probabilities in a row of the transition matrix do not add up to \(1\)

        AssertionError
            if the initial distribution vector and the transition matrix have different number of states
        '''

        assert np.all(init_distr >= 0), 'Probabilities can\'t be negative.'
        assert np.all(trans_mat >= 0), 'Probabilities can\'t be negative.'
        assert np.all(init_distr <= 1), 'Probabilities can\'t be grater than 1.'
        assert np.all(trans_mat <= 1), 'Probabilities can\'t be grater than 1.'

        assert len(init_distr.shape) == 1, 'The initial distribution vector should be a 1D array.' 
        assert np.isclose(init_distr.sum(), 1), 'The probabilities of initial distribution should add up to 1.'

        assert len(trans_mat.shape) == 2, 'The transition matrix should be a 1D array.' 
        assert trans_mat.shape[0] == trans_mat.shape[1], 'Transition matrix should be a square matrix.'
        for row in trans_mat: assert np.isclose(row.sum(), 1), 'The probabilities of all transitions from a state should add up to 1.'

        assert init_distr.shape[0] == trans_mat.shape[0], 'The components of the chain should have the same number of states.'

    @property
    def _zero_mask(self):

        '''
        ### Creates a zero matrix of the same shape the transition matrix has.

        Returns
        -------
        zero_mat : `numpy.ndarray` of shape (`n_states`, `n_states`)
            zero matrix of the transition matrix shape
        '''

        zero_mat = np.zeros(self.shape)
        return zero_mat
    
    @property
    def _all_ones_mask(self):

        '''
        ### Creates a matrix of ones of the same shape the transition matrix has.

        Returns
        -------
        ones_mat : `numpy.ndarray` of shape (`n_states`, `n_states`)
            matrix of ones of the transition matrix shape
        '''

        ones_mat = np.ones(self.shape)
        return ones_mat
    
    def _vertex_image(self, vertices: List[int]) -> List[int]:

        '''
        ### Finds the image of the given set ot vertices (i.e. states).
        Image of a state `i` is a set of all states such that there is a one-step path from `i` to them.
        Image of a set of states contains all states one-step reachable from one of the states in this set.

        Returns
        -------
        image : list[int]
            image of the given set of vertices (i.e. states)

        Raises
        ------
        AssertionError
            if there is an unknown to the system state in `vertices`

        Notes
        -----
        Formally, image is defined as follows:

        $$
        Г(i) = {j | a[i, j] = 1}
        $$

        $$
        Г(B) = union(Г(i)), i ∈ B
        $$

        where \(Г(i)\) / \(Г(B)\) is image of vertex / set of vertices, \((a[i, j])\) is the adjacency matrix
        '''

        states = set(range(self.n_states))
        assert set(vertices).issubset(states), f'The system does not have the following states: {set(vertices) - states}'

        _, image = np.where(self.trans_mat[vertices])
        image = list(set(image))
        image.sort() # set sorts it weirdly
        return image

    #endregion

    #region Iterations

    def trans_insteps(self, n_steps: int, use_canonical=True) -> np.ndarray:

        '''
        ### Calculates matrix of transitions in `n_steps` steps.
        Conceptually, that represents the probabilities that having begun in the state \(i\), the system
        will end up in the state \(j\) having made `n_steps` steps.

        Parameters
        ----------
        n_steps : `int`
            number of steps, where on each step, a transition takes place

        use_canonical : `bool`
            whether to use canonical numbering of the transition matrix

        Returns
        -------
        trans_mat_n : `numpy.ndarray` of shape (`n_states`, `n_states`)
            matrix of probabilities of transitions between states in `n_steps` steps,
            where `trans_mat_n[i, j]` is probability of transition from the state \(i\) to the state \(j\)
            in exactly `n_steps` steps (i.e. `n_steps` transitions)

        Notes
        -----
        Matrix of transitions in `n` step is calculated as follows:

        $$
        P^{(n)} = P^n
        $$

        where \(P(n)\) is matrix of transitions in \(n\) steps, \(n\) is the number of steps,
        \(P\) is matrix of transitions in \(1\) step (i.e. \(P^{(1)}\))

        NB! The rows of the matrix should add up to \(1\) as from each step, the system must be able to go
        to some state anyway, so the sum of the probabilities of the transitions will be 1.
        '''

        matrix = self.trans_mat if not use_canonical else self.trans_mat.canonical
        trans_mat_n = np.linalg.matrix_power(matrix, n=n_steps)
        return trans_mat_n
    
    def states_insteps(self, n_steps: int, use_canonical=True) -> np.ndarray:

        '''
        ### Calculates state distribution vector in `n_steps` steps.
        Conceptually, that represents the probabilities of system to be in the state \(j\) after `n_steps` steps.

        Parameters
        ----------
        n_steps : `int`
            number of steps, where on each step, a transition takes place

        use_canonical : `bool`
            whether to use canonical numbering of the state distribution vector

        Returns
        -------
        state_distr_n : `numpy.ndarray` of shape (`n_states`, )
            state probability distribution in `n_steps` steps (i.e. `n_steps` transitions)

        Notes
        -----
        State distribution in `n` step is calculated as:

        $$
        \pi^{(n)} = \pi^{(0)}P^{(n)}
        $$

        where \(\pi^{(n)}\) is state probability distribution in \(n\) steps,
        \(\pi^{(0)}\) is the initial state distribution

        NB! The vector should add up to \(1\) as the system will be at some state anyway, so the sum
        of the probabilities of all the states will be \(1\).
        '''

        init_distr = self.init_distr if not use_canonical else self.init_distr.canonical
        state_distr_n = np.matmul(self.init_distr, self.trans_insteps(n_steps=n_steps, use_canonical=use_canonical))
        return state_distr_n

    def first_trans_insteps(self, n_steps: int, use_canonical=True) -> np.ndarray:

        '''
        ### Calculates matrix of first transitions in `n_steps` steps.
        Conceptually, that represents the probabilities that having begun in the state \(i\), the system
        will reach the state \(i\) in `n_steps` steps for the first (!) time.

        Parameters
        ----------
        n_steps : `int`
            number of steps, where on each step, a transition takes place

        use_canonical : `bool`
            whether to use canonical numbering of the state distribution vector

        Returns
        -------
        first_trans_mat : `numpy.ndarray` of shape (`n_states`, `n_states`)
            matrix of probabilities of first transitions between states,
            where `first_trans_mat[i, j]` is probability that the system 
            will reach the `j`th state from the `i`th state for the first (!) time in exactly `n_steps` steps

        Notes
        -----
        Matrix of first transitions in `n` step is calculated recursively as follows:

        $$
        \\hat P^{(n)} = (\\hat p_{ij}^{(n)}) | \\hat p^{(n)}_{ij} = \\sum _{k \\neq j}p_{ik}\\hat p^{(n - 1)}_{kj}
        $$
        
        where \(\\hat P^{(n)}\) is matrix of first transitions in \(n\) steps, \(n\) is the number of steps,
        \((p_{ik})\) is the transition matrix (i.e. \(P(1)\)).

        In other words, the probability to get to the state \(j\) from the state \(i\) in \(n\) steps for the first time
        equals the sum of probabilities to get from \(i\) to any state \(h\) (which is not \(j\)) multiplied by
        probability to get to \(j\) from \(h\) for the first time in \(n - 1\) steps.

        NB! Note that this matrix's rows should not add up to \(1\). With a large enough number of steps,
        this matrix might become a zero matrix as it'll be impossible to get to some state for the first time
        in such a big amount of steps.
        '''

        trans_mat = self.trans_mat if not use_canonical else self.trans_mat.canonical
        first_trans_mat = trans_mat.copy() # `Phat(1)`
        for _ in range(n_steps - 1): # -1 because on each step we'll add the probability to finally reach the state
            # calculate recursively: on each step `m`, fill out matrix `Phat(m)`
            local_step = self._zero_mask
            for i in range(self.n_states):
                for j in range(self.n_states):
                    # from `i` to `j` the first time
                    not_j_states = np.delete(trans_mat[i], j) # now we should reach `j` from a not-`j` state so skip it
                    not_j_indices = np.delete(range(self.n_states), j)
                    # finally reach `j` from a not-`j` state: probabilities of path to a not-`j` state * to go from the not-`j`
                    # state to `j` for the first time in `m` - 1 steps (we refer to `phat` for that)
                    probs = not_j_states * first_trans_mat[not_j_indices, j]
                    prob = probs.sum() # all the paths from `i` to `j` where `j` is reached only at the very end
                    local_step[i, j] = prob
            first_trans_mat = local_step # now we'll refer to Phat(m) at the next step 
        return first_trans_mat
    
    def first_trans_nolaterthan(self, n_steps: int, use_canonical=True) -> np.ndarray:

        '''
        ### Calculates matrix of first transitions in no more than `n_steps` steps.
        Conceptually, that represents the probabilities that having begun in the state \(i\), the system
        will reach the state \(j\) in `n_steps` or less steps for the first (!) time.

        Parameters
        ----------
        n_steps : `int`
            number of steps, where on each step, a transition takes place

        use_canonical : `bool`
            whether to use canonical numbering of the state distribution vector

        Returns
        -------
        first_trans_mat_nolater : `numpy.ndarray` of shape (`n_states`, `n_states`)
            matrix of probabilities of first transitions between states,
            where `first_trans_mat_nolater[i, j]` is probability that the system 
            will reach the `j`th state from the `i`th state for the first (!) time in `n_steps` steps or less

        Notes
        -----
        Matrix of first transitions in no more than `n` step is calculated as sum of the 
        matrices of first transitions in exactly \(1\), \(2\), ..., `n` steps:

        $$
        \\hat P^{(\\leq n)} = \\sum _{t=1}^{n}\\hat P^t
        $$

        where \(\\hat P^{(\leq n)}\) is matrix of first transitions in \(n\) steps or less,
        \(\\hat P^{(n)}\) is matrix of first transitions in \(n\) steps, \(n\) is the number of steps.

        NB! Note that this matrix's rows should not add up to \(1\). With a large enough number of steps,
        the sum might be more that \(1\) because in a little number of steps,
        it might be possible to get to several states for the first time with a high probability.
        '''

        first_trans_mat_nolater = self._zero_mask
        for n in range(1, n_steps): first_trans_mat_nolater += self.first_trans_insteps(n_steps=n, use_canonical=use_canonical)
        return first_trans_mat_nolater

    #endregion

    #region Reachability

    def _adjacency_mat(self) -> np.ndarray:

        '''
        ### Builds adjacency matrix of the chain.
        Adjacency matrix is a matrix that indicates existence of one-step paths between states,
        where a one-step path from the state \(i\) to the state \(j\) exists 
        if there a non-zero probability transition between them.

        Returns
        -------
        adjacency_mat : `March._matrix` of shape (`n_states`, `n_states`)
            boolean-like matrix indicating one-step paths between states,
            where `adjacency_mat[i, j]` is \(1\) if the probability of tranition 
            from the state `i` to the state `j` is not zero, else \(0\)

        Notes
        -----
        Adjacency matrix is calculated as shown below:

        $$
        A = (a_{ij}) | a_{ij} =
        \\begin{cases}
            1, & p_{ij} > 0     \\\\
            0, & p_{ij} = 0
        \\end{cases}
        $$

        where \(A\) is adjacency matrix, \((p_{ij})\) is transition matrix.
        '''

        adjacency_mat = self._all_ones_mask
        adjacency_mat *= (self.trans_mat > 0)
        return adjacency_mat.astype(int)
    
    def _reachability_mat(self) -> np.ndarray:

        '''
        ### Builds reachability matrix following the Warshall's Algorithm. 
        Reachability matrix is a matrix that indicates existence of paths between states,
        where a path from the state \(i\) to the state \(j\) exists if there a sequence 
        of non-zero probability transitions between states such that it leads from \(i\) to \(j\)
        (e.g. \(i\) → \(h\) → \(k\) → ... → \(j\)).

        Returns
        -------
        reachability_mat : `March._matrix` of shape (`n_states`, `n_states`)
            boolean-like matrix indicating paths between states,
            where `reachability_mat[i, j]` is \(1\) if there is a path 
            from the state `i` to the state `j`, else \(0\)

        Notes
        -----
        Reachability matrix is calculated as shown below:

        $$
        D = (d_{ij}) | d_{ij} =
        \\begin{cases}
            1, & \\exists i → j        \\\\
            0, & \\neg \exists i → j
        \\end{cases}
        $$

        where \(D\) is reachability matrix, \(→\) marks a path.
        '''

        reachability_mat = self._zero_mask
        curr_adj_mat = self.adjacency_mat.copy() # start with the adjacency matrix
        for state in range(0, self.n_states):
            # states from where `state` is accessible in the current adjacency matrix
            adj_states, = np.where(curr_adj_mat[:, state])
            # for each of those states, update the row as such:
            # `A[i] := A[i] v A[state]`, where `i` ∈ `adj_states`
            curr_adj_mat[adj_states] |= curr_adj_mat[state]
        reachability_mat = curr_adj_mat
        return reachability_mat

    def _reachability_mat_tr(self) -> np.ndarray:

        '''
        ### Builds transposed reachability matrix. 
        Reachability matrix is a matrix that indicates existence of inverse paths between states,
        where a path from the state \(i\) to the state \(j\) exists if there a sequence 
        of non-zero probability transitions between states such that it leads from \(i\) to \(j\)
        (e.g. \(i\) → \(h\) → \(k\) → ... → \(j\)).

        Returns
        -------
        reachability_mat_tr : `March._matrix` of shape (`n_states`, `n_states`)
            boolean-like matrix indicating paths between states,
            where `reachability_mat_tr[i, j]` is \(1\) if there is a path 
            from the state `j` to the state `i`, else \(0\)

        Notes
        -----
        Transposed reachability matrix is calculated as shown below:

        $$
        D^T = (d^T_{ij}) | d^T_{ij} =
        \\begin{cases}
            1, & \\exists j → i        \\\\
            0, & \\neg \exists j → i
        \\end{cases}
        $$

        where \(D^T\) is transposed reachability matrix, \(→\) marks a path.
        '''

        reachability_mat_tr = self.reachability_mat.T
        return reachability_mat_tr
    
    def _communication_mat(self) -> np.ndarray:

        '''
        ### Builds communication matrix. 
        Communication matrix is a matrix that indicates existence of bidirectional paths between states,
        where a bidirectional path between the states \(i\) and \(j\) exists if there a sequence 
        of non-zero probability transitions from \(i\) to \(j\) (e.g. \(i\) → \(h\) → \(k\) → ... → \(j\))
        and from \(j\) to \(i\) (not exactly the same way back).

        Returns
        -------
        communication_mat : `March._matrix` of shape (`n_states`, `n_states`)
            boolean-like matrix indicating bidirectional paths between states,
            where `communication_mat[i, j]` is \(1\) if there is a path 
            from the state `j` to the state `i` and from `j` to `i`, else \(0\)

        Notes
        -----
        Communication matrix is calculated as shown below:

        $$
        C = (c_{ij}) = D \\times D^T | c_{ij} = 
        \\begin{cases}
            1, & d_{ij} = d_{ji} = 1        \\\\
            0, & \\neg (d_{ij} = d_{ji} = 1)
        \\end{cases}
        $$

        where \(C\) is communication matrix, \(D\) is reachability matrix,
        \(D^T\) is transposed reachability matrix, \(→\) marks a path.
        '''

        communication_mat = self.reachability_mat & self.reachability_mat_tr
        return communication_mat
    
    def _communication_mat_comp(self) -> np.ndarray:

        '''
        ### Builds communication matrix complement. 
        Communication matrix complement is a matrix that indicates nonexistence of bidirectional paths between states,
        where a bidirectional path between the states \(i\) and \(j\) exists if there a sequence 
        of non-zero probability transitions from \(i\) to \(j\) (e.g. \(i\) → \(h\) → \(k\) → ... → \(j\))
        and from \(j\) to \(i\) (not exactly the same way back).

        Returns
        -------
        communication_mat_comp : `March._matrix` of shape (`n_states`, `n_states`)
            boolean-like matrix indicating absence of bidirectional paths between states,
            where `communication_mat[i, j]` is \(1\) if there is no path 
            from the state `j` to the state `i` and from `j` to `i`, else 0

        Notes
        -----
        Communication matrix complement is calculated as follows:

        $$
        \\overline C = (\\overline c_{ij}) | \\overline c_{ij} = 
        \\begin{cases}
            1, & c_{ij} = 0     \\\\
            0, & c_{ij} = 1
        \\end{cases}
        $$

        where \(\\overline C\) is communication matrix complement, \(c_{ij}\) is communication matrix.
        '''

        communication_mat = self.communication_mat.copy().astype(bool)
        communication_mat_comp = np.invert(communication_mat).astype(int)
        return communication_mat_comp
    
    def _classification_mat(self) -> np.ndarray:

        '''
        ### Builds classification matrix. 
        Classification matrix is a step for building extended classification matrix.

        Returns
        -------
        classification_mat : `March._matrix` of shape (`n_states`, `n_states`)
            boolean-like matrix where `classification_mat[i, j]` is \(1\) if there is a path 
            from the state `i` to the state `j` but states `j` and `i` do not communicate, 0 otherwise

        Notes
        -----
        Classification matrix is calculated as follows:

        $$
        T = D \\times \\overline C = (t_{ij}) | t_{ij} = 
        \\begin{cases}
            1, & d_{ij} = \\overline c_{ij} = 1      \\\\
            0, & \\neg (d_{ij} = \\overline c_{ij} = 1)
        \\end{cases}
        $$

        where \(T\) is classification matrix, \(d_{ij}\) is reachability matrix, 
        \((\\overline c_{ij})\) is communication matrix complement.
        '''

        classification_mat = self.reachability_mat & self.communication_mat_comp
        return classification_mat
    
    def _classification_mat_ext(self) -> np.ndarray:

        '''
        ### Builds classification matrix extension. 
        Classification matrix extension indicates essential and inessential states.
        Essential states are the ones that communicate with each (!) state they have a path to;
        respectively, the ineccential are the ones for which this statement is not true.

        Returns
        -------
        classification_mat_ext : `March._matrix` of shape (`n_states`, `n_states` + 1)
            boolean-like matrix where the last column indicates essentiality of states:
            if classification_mat_ext[`i`, `n_states` + 1] equals 1, the state `i`
            is inessential, if 0, is essential

        Notes
        -----
        Classification matrix extention is defined below:

        $$
        T_{ext} = T|(t_{i(N + 1)}) | t_{i(N + 1)} = 
        \\begin{cases}
            1, & \\exists j t_{ij} = 1        \\\\
            0, & \\neg \\exists j t_{ij} = 1
        \\end{cases}
        $$

        where \(T_{ext}\) is classification matrix extension, \(T\) is classification matrix,
        \(N\) is the number of states in the system.
        '''

        classification_mat = self.classification_mat.copy()
        extension = np.any(classification_mat, axis=1).astype(int)
        classification_mat_ext = np.concatenate((classification_mat, extension.reshape(-1, 1)), axis=1)
        return classification_mat_ext

    #endregion

    #region Structure

    def _build_structure(self) -> None:

        '''
        ### Build structure of the chain (into `structure` property).
        That includes:

        * Determining essential and inessential states:
        Essential state is such a state \(i\) that communicates with every (!) state \(j\)
        it has a path to; respectively, a state is inessential if that is not true.

        * Building matrix of equivalency classes:
        In each Markov system, the set of essential states can be uniquely split into 
        classes of equivalency. In each equivalency class, all essential states
        communicate with each other and each class is closed, which means
        once the system got into an essential state in class \(EC_i\), it will never leave that class
        and all the transitions will be only between states within this class.

        * Finding cyclic subclasses of equivalence classes:
        Each equivalence classes has a period \(d\): it equals the GCD of the lengths
        of all the contours going through a state \(i\) belonging to the equivalence class.
        Contour is a cyclic path from a state to itself (e.g. \(i → h → k → ... → i\)), 
        its length is the number of the transitions within it.
        It is proven that all the states of an equivalence class have the same period \(d\).
        If \(d\) equals 1, the equivalence class is acyclic.
        If it is \(\geq\) 2, it can be split into \(d\) cyclic subclasses. Cyclic subclass \(C_r\) of equivalence class \(EC_n\)
        is a subset of states of \(EC_n\) such that 
        
        * the states are not adjacent (i.e. there is no one-step path between them);
        * from any of the states there is a one-step path only to a state of the next cyclic subclass \(C_{r + 1}\);
        * the states belong to an only cyclic subclass;
        * the equivalence class is fully distributed between cyclic subclasses (no states left outside of the subclasses).

        That means the cyclic subclasses are closed and form a cycle themselves:

        \(C_0 → C_1 → C_2 → ... → C_{d - 1} → C_0\)

        * Canonical numbering of the states in the system structure (transformed chain components are built separately):
        Canonical numbering is a numbering that goes in strict order through each state in each cyclic subclass
        of each equivalency class, followed only then by inessential states. That allows to transform the transition matrix
        into a block matrix that represents probabilities of transition between cyclic subclasses.

        Returns
        -------
        equivalency_cls_mat : `numpy.ndarray` of shape (`n_classes`, `n_essential`)
            matrix showing belonging of each essential (!) state to its equivalence class;
            `equivalency_cls_mat[i, j]` indicates that the state `j` is in `i`th equivalence class

        Notes
        -----
        Essentiality of states if defined as follows:
        $$
        i \\in S_e \\iff 
        \\forall j: \\quad i → j \\Longrightarrow j → i
        $$
        $$
        i \\in S_{ie} \\iff 
        \\neg \\forall j: \\quad i → j \\Longrightarrow j → i
        $$

        Where \(S_e\) is the set of essential states, \(S_{ie}\) is the set of inessential states.


        Matrix of equivalency classes is defined as follows:

        $$
        K = (k_{ij}) | k_{it} \\Longleftrightarrow i \\in EC_t
        $$

        where \(K\) is equivalency matrix

        Formally, the period of an equivalency class is found as:

        $$
        d = \\text{GCD}(|(i → i)_1|, ..., |(i → i)_r|)
        $$
            
        where \((i → i)\) is a contour, \(|(i → i)_1|\) is a length of a contour

        If \(d\) equials 1, the equivalence class is acyclic, which means it has no cyclic subclasses.

        
        Canonical numbering goes as follows:

        \(EC_1\):

        \(C_0: \\quad 1, \\quad 2, \\quad 3, \\quad ..., \\quad r_0\)

        \(C_1: \\quad r_0 + 1, \\quad r_0 + 2, \\quad ..., \\quad r_1\)

        \(...\)

        \(C_{d - 1}: r_0 + r_1 + ... + r_{d - 1} + 1, \\quad r_0 + r_1 + ... + r_{d - 1} + 2, 
        \\quad ..., \\quad r_0 + r_1 + ... + r_{d - 1} + r_{d} = N_{EC_1}\)

        \(EC_2\): (same way)

        \(...\)

        \(EC_L\): (same way)

        \(S_{ie}: \\quad N_e + 1, \\quad N_e + 2, \\quad ..., \\quad N_e + N_{ie} = N\)

        where \(EC_k\) is equivalence class \(k\), \(C_s\) is a cyclic subclass \(s\) of the equivalence class,
        \(N_{EC_k}\) is number of states in equivalence class \(k\), \(N_e\) is number of essential states in the system,
        \(N\) is number of states in the system.
        '''

        # NB! We keep all the data pieces as lists as the dataclasses for them are immutable after initialization
        can_id = 0 # for canonical numbering
        # collections for the whole chain
        states_, es_states_, ines_states_ = [], [], []
        cyclic_subclasses_ = []
        equivalency_classes_ = []

        # determine essential and inessential states
        extention = self.classification_mat_ext[:, -1]
        essential_states, = np.where(extention == 0)
        inessential_states, = np.where(extention)

        # equivalency classification matrix
        communication_mat = self.communication_mat.copy()
        # delete rows and columns of all inessential states
        communication_mat = np.delete(communication_mat, inessential_states, axis=0)
        communication_mat = np.delete(communication_mat, inessential_states, axis=1)
        # the resulting matrix will fall apart into groups of repeating rows
        # after removing dublicates we'll get our equivalency classes matrix
        equivalency_cls_mat, indices = np.unique(communication_mat, return_index=True, axis=0)
        equivalency_cls_mat = equivalency_cls_mat[indices.argsort()]

        # in the equivalency classification matrix, the indices of the states
        # might not correspond to the actual state indexing (since we delete columns),
        # which means we need a mapping between the index of a state in the matrix
        # and the actual essential state id 
        mapping = {
            number: state for state, number 
            # `i`th index in the matrix is the `i`th essential (!) state 
            in zip(essential_states, range(len(essential_states)))
        }
        for cl_id, row in enumerate(equivalency_cls_mat):
            class_states, = np.where(row)
            equivalency_class = [mapping[class_state] for class_state in class_states]

            # collections for the equivalence class
            cls_states_ = []
            subclasses_ = []

            # determine cyclic subclasses
            candidates = [[equivalency_class[0]]] # start with any state in the equivalence class
            while True:

                image = self._vertex_image(candidates[-1]) # find its image
                if image == candidates[0]: break # cycle is closed
                candidates.append(image) # proceed otherwise

                to_update, to_pop = [], []
                for ind, prev_cand in enumerate(candidates[:-1]): # check if any of previous candidates overlap with the current one
                    if set(prev_cand) & set(candidates[-1]):
                        to_update.extend(prev_cand)
                        to_pop.append(ind)
                candidates[-1].extend(to_update) # merge if overlap
                candidates[-1] = list(set(candidates[-1]))
                candidates[-1].sort() # set sorts it weirdly
                candidates = [
                    candidate for j, candidate in enumerate(candidates)
                    if not j in to_pop
                ] # remove old candidates, now they are merged into a new one
            cyclic_subclasses = candidates

            # fill in the structure: since we will add 0th to `r`th state of every
            # 0th to `m`th cyclic subclass of each 0th to `s`th equivalency class,
            # the order of appending will follow the canonical numbering!
            for subcl_id, state_ids in enumerate(cyclic_subclasses):
                
                # collection for the cyclic subclass
                subcl_states_ = []

                for state_id in state_ids:
                    state = EssentialState(
                        id=state_id,
                        canonical_number=can_id,
                        equivalency_class=cl_id,
                        cyclic_subclass=subcl_id
                    )

                    # update cyclic subclass
                    subcl_states_.append(state)
                    can_id += 1
                
                subclass = CyclicSubclass(
                    id=subcl_id,
                    class_id=cl_id,
                    states=subcl_states_
                )

                # update equivalency class
                cls_states_.extend(subcl_states_)
                subclasses_.append(subclass)

            d = len(subclasses_)
            eq_class = EquivalenceClass(
                id=cl_id,
                d=d,
                states=cls_states_,
                cyclic_subclasses=subclasses_ if d > 1 else None
            )

            # update chain structure
            cyclic_subclasses_.extend(subclasses_)
            equivalency_classes_.append(eq_class)
            states_.extend(cls_states_); es_states_.extend(cls_states_)

        # finally add inessential states
        for state_id in inessential_states:
            state = InessentialState(
                id=state_id,
                canonical_number=can_id
            )
            states_.append(state); ines_states_.append(state)
            can_id += 1

        structure = ChainStructure(
            states=states_,
            essential_states=es_states_,
            inessential_states=ines_states_,
            equivalency_classes=equivalency_classes_,
            cyclic_subclasses=cyclic_subclasses_
        )

        self.structure = structure
        return equivalency_cls_mat
  
    #endregion

    #region Classification

    def _classify(self) -> None:

        '''
        ### Classifies the chain (into `properties` property).
        Markov Chain classification:

        1. \(S\) = \(Se\) = \(EC\) ⇒ ergodic / irreducible Markov Chain

            1.1. \(EC\) is regular ⇒ regular irreducible Markov Chain

            \(\quad\) 1.1.1. \(EC\) is absorbing ⇒ absorbing irreducible Markov Chain

            1.2. \(EC\) is cyclic ⇒ cyclic irreducible Markov Chain

        2. \(Sie ≠ ∅\) ⇒ reducible Markov Chain

            2.1. a single \(EC\) ⇒ monoergodic Markov Chain

            \(\quad\)2.1.1. the single \(EC\) is regular ⇒ regular monoergodic Markov Chain

            \(\qquad\) 2.1.1.1. the single \(EC\) is absorbing ⇒ absorbing monoergodic Markov Chain

            \(\quad\) 2.1.2. the single \(EC\) is cyclic ⇒ cyclic monoergodic Markov Chain

            2.2. more than one \(EC\)s ⇒ polyergodic Markov Chain

            \(\quad\) 2.2.1. all the \(EC\)s are regular ⇒ regular polyergodic Markov Chain

            \(\qquad\) 2.2.1.1. all the \(EC\)s are absorbing ⇒ absorbing polyergodic Markov Chain

            \(\quad\) 2.2.2. all the \(EC\)s are cyclic ⇒ cyclic polyergodic Markov Chain

            \(\quad\) 2.2.3. there are both regular and cyclic \(EC\)s ⇒ mixed polyergodic Markov Chain,

        where \(S\) is set of states, \(Se\) is set of essential states, \(Sie\) is set of inessential states,
        \(EC\) stands for equivalence class.

        Returns
        -------
        structure : `ChainProperties`
            class containing attributes for chain properties

        Notes
        -----
        Note that in our classification we won't just return a single label of the class, but rather
        use a more "linguistic" notation: we will define a set of properties, each of which can be 
        either `True`, or `False`, or `None` (if not applicable). Thus, we will be able to 
        split the space of classes uniquely, yet be able to access chain features in a more convenient way.

        Reminder: a \(EC\) is regular if it has a period \(d\) = 1, i.e. has no cyclic subclasses, otherwise cyclic.
        Reminder: a \(EC\) is absorbing if it contains a single state.
        
        The properties are:

        * `reducible` (if there are inessential states)
        * `polyergodic` (if there are more than one `EC`)
        * `regular` (cyclic if `False`, mixed if `None`)
        * `absorbing` (may be `True` only if regular)
        * `strong_convergence` (if the matrix strongly converges to a final matrix in a great number of steps)

        The matrix converges strongly if it's regular, weakly otherwise.

        Thus, we'll have a "linguistic-like" notation, e.g. a cyclic polyergodic Markov Chain
        could be notated as ```[+reducible][+polyergodic][-regular][-absorbing][-strong_convergence]```.
        '''

        reducible = None
        polyergodic = None
        regular = None
        absorbing = None
        strong_convergence = None

        if not self.n_inessential_states: # irreducible
            
            reducible = False
            polyergodic = False # it has a single equivalency class then
            period = self.structure.equivalency_classes[0].d
            len_eq_class = len(self.structure.equivalency_classes[0])
            regular = period == 1
            absorbing = len_eq_class == 1 if regular else False

        else: # reducible
            
            reducible = True

            if self.n_equivalency_classes == 1: # monoergodic

                polyergodic = False
                period = self.structure.equivalency_classes[0].d
                len_eq_class = len(self.structure.equivalency_classes[0])
                regular = period == 1
                absorbing = len_eq_class == 1 if regular else False

            else: # polyergodic

                polyergodic = True
                periods = [eq_class.d for eq_class in self.structure.equivalency_classes]
                lens_eq_class = [len(eq_class) for eq_class in self.structure.equivalency_classes]
                if np.all(np.array(periods) == 1): regular = True
                elif np.all(np.array(periods) > 1): regular = False
                # else: regular = None # mixed
                absorbing = np.mean(lens_eq_class) == 1 if regular else False

        strong_convergence = regular == True # might be `None` is mixed so compare to `True`

        self.properties = ChainProperties(
            reducible=reducible,
            polyergodic=polyergodic,
            regular=regular,
            absorbing=absorbing,
            strong_convergence=strong_convergence
        )

    #endregion

    def to_graph(self, use_canonical=True):

        '''
        ### Shows the graph on the transition matrix.

        Parameters
        ----------
        use_canonical : `bool`
            whether to use canonical numbering

        Returns
        -------
        graph : `graphviz.Digraph`
            graph
        '''

        matrix = self.trans_mat.canonical if use_canonical else self.trans_mat.original

        graph = graphviz.Digraph()
        for i, row in enumerate(matrix):
            for j, value in enumerate(row):
                if value == 0: continue
                graph.edge(
                    tail_name=str(i),
                    head_name=str(j),
                    label=str(value),
                    arrowhead='empty',
                    arrowsize='0.6',
                    fontsize='12',
                    shape='circle'
                )
        graph.attr(
            layout='dot',
            size='12.0',
            rankdir='LR'
        )
        graph.node_attr['shape'] = 'circle'
        return graph

    def __len__(self) -> int:
        return self.n_states

    def __repr__(self) -> str:
        return f'{self.properties.type_label} {repr(self.structure)}\n\n{repr(self.properties)}'
    
    def __str__(self) -> str:
        return self.__repr__()


#region Imports

import unittest
import numpy as np

from marchie import MarChie
from marchie.marchie import *

#endregion


#region Tests

class MarChieTest(unittest.TestCase):

    '''
    ### Tests features of a Markov Chain: reachability, communication etc.
    '''

    #region Check Chain Tests

    def test_check_marchie_0(self):

        trans_mat = np.array([
            [0.5,   0.5],
            [1,     0  ]
        ])
        marchie = MarChie(trans_mat=trans_mat)

        self.assertIsInstance(marchie, MarChie)

    def test_check_marchie_1(self):
        
        trans_mat = np.array([
            [0.5,   0.5],
            [1,     0  ]
        ])
        init_distr = np.array([0.3, 0.7])
        marchie = MarChie(init_distr=init_distr, trans_mat=trans_mat)
        
        self.assertIsInstance(marchie, MarChie)

    def test_check_marchie_2(self):
        
        trans_mat = np.array([
            [0.5,   0.2], # < 1
            [1,     0  ]
        ])

        with self.assertRaises(AssertionError): MarChie(trans_mat=trans_mat)

    def test_check_marchie_3(self):
        
        trans_mat = np.array([
            [0.5,   0.5],
            [1,     0.1] # > 1
        ])

        with self.assertRaises(AssertionError): MarChie(trans_mat=trans_mat)

    def test_check_marchie_4(self):
        
        trans_mat = np.array([
            [0.5,   0.5],
            [1,     0  ]
        ])
        init_distr = np.array([0.3, 0.3]) # < 1
        
        with self.assertRaises(AssertionError): MarChie(init_distr=init_distr, trans_mat=trans_mat)

    def test_check_marchie_5(self):
        
        trans_mat = np.array([
            [0.5,   0.5],
            [1,     0  ]
        ])
        init_distr = np.array([0, 2]) # > 1
        
        with self.assertRaises(AssertionError): MarChie(init_distr=init_distr, trans_mat=trans_mat)

    def test_check_marchie_6(self):
        
        trans_mat = np.array([
            [2,   -1], # invalid probs
            [1,    0]
        ])
        init_distr = np.array([0, 1])
        
        with self.assertRaises(AssertionError): MarChie(init_distr=init_distr, trans_mat=trans_mat)

    def test_check_marchie_7(self):
        
        trans_mat = np.array([
            [0.5,  0.5],
            [1,    0  ]
        ])
        init_distr = np.array([0, -1]) # invalid probs
        
        with self.assertRaises(AssertionError): MarChie(init_distr=init_distr, trans_mat=trans_mat)

    def test_check_marchie_8(self):
        
        trans_mat = np.array([
            [[0.5,   0.5], # 3D matrix
             [1,     0 ]]
        ])

        with self.assertRaises(AssertionError): MarChie(trans_mat=trans_mat)

    def test_check_marchie_9(self):
        
        trans_mat = np.array([
            [0.5,   0.5],
            [1,     0  ]
        ])
        init_distr = np.array([[0, 1]]) # 2D matrix

        with self.assertRaises(AssertionError): MarChie(init_distr=init_distr, trans_mat=trans_mat)

    def test_check_marchie_10(self):
        
        trans_mat = np.array([
            [0.5,   0.5], # not a square matrix
            [1,     0  ],
            [0.3,   0.7]
        ])

        with self.assertRaises(AssertionError): MarChie(trans_mat=trans_mat)

    def test_check_marchie_11(self):
        
        trans_mat = np.array([
            [0.5,   0.2,   0.3], # not a square matrix
            [0.9,   0,     0.1]
        ])

        with self.assertRaises(AssertionError): MarChie(trans_mat=trans_mat)

    def test_check_marchie_12(self):
        
        trans_mat = np.array([
            [0.5,   0.2], # different number of states
            [0.9,   0  ]
        ])
        init_distr = np.array([0, 0.4, 0.6])

        with self.assertRaises(AssertionError): MarChie(init_distr=init_distr, trans_mat=trans_mat)

    #endregion

    #region Communication Tests

    def test_adjacency_mat_0(self):
        
        trans_mat = np.array([
            [1,     0,     0  ],
            [0.8,   0.2,   0  ],
            [0.3,   0.5,   0.2]
        ])
        marchie = MarChie(trans_mat=trans_mat)

        expected_adjacency_mat = np.array([
            [1,     0,     0  ],
            [1,     1,     0  ],
            [1,     1,     1  ]
        ])
        actual_adjacency_mat = marchie.adjacency_mat.original

        self.assertTrue(
            np.array_equal(expected_adjacency_mat, actual_adjacency_mat)
        )
        self.assertIs(actual_adjacency_mat.dtype, np.dtype(int))

    def test_adjacency_mat_1(self):
        
        trans_mat = np.array([
            [1e-10,    1 - 1e-10,    0       ],
            [0,        1e-5,         1 - 1e-5],
            [0,        1,            0       ]
        ])
        marchie = MarChie(trans_mat=trans_mat)

        expected_adjacency_mat = np.array([
            [1,     1,     0],
            [0,     1,     1],
            [0,     1,     0]
        ])
        actual_adjacency_mat = marchie.adjacency_mat.original

        self.assertTrue(
            np.array_equal(expected_adjacency_mat, actual_adjacency_mat)
        )
        self.assertIs(actual_adjacency_mat.dtype, np.dtype(int))

    def test_reachability_mat_0(self):
        
        trans_mat = np.array([
            [0.01,   0.09,   0.9,    0   ],
            [0.03,   0.07,   0,      0.9 ],
            [0.7,    0,      0.21,   0.09],
            [0,      0.7,    0.03,   0.27]
        ])
        marchie = MarChie(trans_mat=trans_mat)

        expected_reachability_mat = np.array([
            [1,     1,     1,     1],
            [1,     1,     1,     1],
            [1,     1,     1,     1],
            [1,     1,     1,     1]
        ])
        actual_reachability_mat = marchie.reachability_mat.original

        self.assertTrue(
            np.array_equal(expected_reachability_mat, actual_reachability_mat)
        )
        self.assertIs(actual_reachability_mat.dtype, np.dtype(int))

    def test_reachability_mat_1(self):
        
        trans_mat = np.array([
            [1,     0,     0  ],
            [0.8,   0.2,   0  ],
            [0.3,   0.5,   0.2]
        ])
        marchie = MarChie(trans_mat=trans_mat)

        expected_reachability_mat = np.array([
            [1,     0,     0  ],
            [1,     1,     0  ],
            [1,     1,     1  ]
        ])
        actual_reachability_mat = marchie.reachability_mat.original

        self.assertTrue(
            np.array_equal(expected_reachability_mat, actual_reachability_mat)
        )
        self.assertIs(actual_reachability_mat.dtype, np.dtype(int))

    def test_reachability_mat_tr_0(self):
        
        trans_mat = np.array([
            [0.01,   0.09,   0.9,    0   ],
            [0.03,   0.07,   0,      0.9 ],
            [0.7,    0,      0.21,   0.09],
            [0,      0.7,    0.03,   0.27]
        ])
        marchie = MarChie(trans_mat=trans_mat)

        expected_reachability_mat_tr = np.array([
            [1,     1,     1,     1],
            [1,     1,     1,     1],
            [1,     1,     1,     1],
            [1,     1,     1,     1]
        ])
        actual_reachability_mat_tr = marchie.reachability_mat_tr.original

        self.assertTrue(
            np.array_equal(expected_reachability_mat_tr, actual_reachability_mat_tr)
        )
        self.assertIs(actual_reachability_mat_tr.dtype, np.dtype(int))

    def test_reachability_mat_tr_1(self):
        
        trans_mat = np.array([
            [1,     0,     0  ],
            [0.8,   0.2,   0  ],
            [0.3,   0.5,   0.2]
        ])
        marchie = MarChie(trans_mat=trans_mat)

        expected_reachability_mat_tr = np.array([
            [1,     1,     1],
            [0,     1,     1],
            [0,     0,     1]
        ])
        actual_reachability_mat_tr = marchie.reachability_mat_tr.original

        self.assertTrue(
            np.array_equal(expected_reachability_mat_tr, actual_reachability_mat_tr)
        )
        self.assertIs(actual_reachability_mat_tr.dtype, np.dtype(int))

    def test_communication_mat_0(self):
        
        trans_mat = np.array([
            [0.55,  0.45,  0],
            [0,     0,     1],
            [1,     0,     0]
        ])
        marchie = MarChie(trans_mat=trans_mat)

        expected_communication_mat = np.array([
            [1,     1,     1],
            [1,     1,     1],
            [1,     1,     1]
        ])
        actual_communication_mat = marchie.communication_mat.original

        self.assertTrue(
            np.array_equal(expected_communication_mat, actual_communication_mat)
        )
        self.assertIs(actual_communication_mat.dtype, np.dtype(int))

    def test_communication_mat_1(self):
        
        trans_mat = np.array([
            [1,     0,     0  ],
            [0.8,   0.2,   0  ],
            [0.3,   0.5,   0.2]
        ])
        marchie = MarChie(trans_mat=trans_mat)

        expected_communication_mat = np.array([
            [1,     0,     0],
            [0,     1,     0],
            [0,     0,     1]
        ])
        actual_communication_mat = marchie.communication_mat.original

        self.assertTrue(
            np.array_equal(expected_communication_mat, actual_communication_mat)
        )
        self.assertIs(actual_communication_mat.dtype, np.dtype(int))

    def test_communication_mat_comp_0(self):
        
        trans_mat = np.array([
            [0.55,  0.45,  0],
            [0,     0,     1],
            [1,     0,     0]
        ])
        marchie = MarChie(trans_mat=trans_mat)

        expected_communication_mat_comp = np.array([
            [0,     0,     0],
            [0,     0,     0],
            [0,     0,     0]
        ])
        actual_communication_mat_comp = marchie.communication_mat_comp.original

        self.assertTrue(
            np.array_equal(expected_communication_mat_comp, actual_communication_mat_comp)
        )
        self.assertIs(actual_communication_mat_comp.dtype, np.dtype(int))

    def test_communication_mat_comp_1(self):
        
        trans_mat = np.array([
            [1,     0,     0  ],
            [0.8,   0.2,   0  ],
            [0.3,   0.5,   0.2]
        ])
        marchie = MarChie(trans_mat=trans_mat)

        expected_communication_mat_comp = np.array([
            [0,     1,     1],
            [1,     0,     1],
            [1,     1,     0]
        ])
        actual_communication_mat_comp = marchie.communication_mat_comp.original

        self.assertTrue(
            np.array_equal(expected_communication_mat_comp, actual_communication_mat_comp)
        )
        self.assertIs(actual_communication_mat_comp.dtype, np.dtype(int))

    #endregion

    #region Classification Tests

    def test_classification_mat_0(self):
        
        trans_mat = np.array([
            [0.55,  0.45,  0],
            [0,     0,     1],
            [1,     0,     0]
        ])
        marchie = MarChie(trans_mat=trans_mat)

        expected_classification_mat = np.array([
            [0,     0,     0],
            [0,     0,     0],
            [0,     0,     0]
        ])
        actual_classification_mat = marchie.classification_mat.original

        self.assertTrue(
            np.array_equal(expected_classification_mat, actual_classification_mat)
        )
        self.assertIs(actual_classification_mat.dtype, np.dtype(int))

    def test_classification_mat_1(self):
        
        trans_mat = np.array([
            [1,     0,     0  ],
            [0.8,   0.2,   0  ],
            [0.3,   0.5,   0.2]
        ])
        marchie = MarChie(trans_mat=trans_mat)

        expected_classification_mat = np.array([
            [0,     0,     0],
            [1,     0,     0],
            [1,     1,     0]
        ])
        actual_classification_mat = marchie.classification_mat.original

        self.assertTrue(
            np.array_equal(expected_classification_mat, actual_classification_mat)
        )
        self.assertIs(actual_classification_mat.dtype, np.dtype(int))

    def test_classification_mat_ext_0(self):
        
        trans_mat = np.array([
            [0.55,  0.45,  0],
            [0,     0,     1],
            [1,     0,     0]
        ])
        marchie = MarChie(trans_mat=trans_mat)

        expected_classification_mat_ext = np.array([
            [0,     0,     0,     0],
            [0,     0,     0,     0],
            [0,     0,     0,     0]
        ])
        actual_classification_mat_ext = marchie.classification_mat_ext.original

        self.assertTrue(
            np.array_equal(expected_classification_mat_ext, actual_classification_mat_ext)
        )
        self.assertIs(actual_classification_mat_ext.dtype, np.dtype(int))

    def test_classification_mat_ext_1(self):
        
        trans_mat = np.array([
            [1,     0,     0  ],
            [0.8,   0.2,   0  ],
            [0.3,   0.5,   0.2]
        ])
        marchie = MarChie(trans_mat=trans_mat)

        expected_classification_mat_ext = np.array([
            [0,     0,     0,     0],
            [1,     0,     0,     1],
            [1,     1,     0,     1]
        ])
        actual_classification_mat_ext = marchie.classification_mat_ext.original

        self.assertTrue(
            np.array_equal(expected_classification_mat_ext, actual_classification_mat_ext)
        )
        self.assertIs(actual_classification_mat_ext.dtype, np.dtype(int))

    def test_equivalency_cls_mat_0(self):
        
        trans_mat = np.array([
            [0.55,  0.45,  0],
            [0,     0,     1],
            [1,     0,     0]
        ])
        marchie = MarChie(trans_mat=trans_mat)

        expected_equivalency_cls_mat = np.array([
            [1,     1,     1]
        ])
        actual_equivalency_cls_mat = marchie.equivalency_cls_mat.original

        self.assertTrue(
            np.array_equal(expected_equivalency_cls_mat, actual_equivalency_cls_mat)
        )
        self.assertIs(actual_equivalency_cls_mat.dtype, np.dtype(int))

    def test_equivalency_cls_mat_1(self):
        
        trans_mat = np.array([
            [1,     0,     0  ],
            [0.8,   0.2,   0  ],
            [0.3,   0.5,   0.2]
        ])
        marchie = MarChie(trans_mat=trans_mat)

        expected_equivalency_cls_mat = np.array([
            [1]
        ])
        actual_equivalency_cls_mat = marchie.equivalency_cls_mat.original

        self.assertTrue(
            np.array_equal(expected_equivalency_cls_mat, actual_equivalency_cls_mat)
        )
        self.assertIs(actual_equivalency_cls_mat.dtype, np.dtype(int))

    def test_equivalency_cls_mat_2(self):
        
        trans_mat = np.array([
            [1,     0,     0,     0,     0,     0,     0  ],
            [0,     0,     1,     0,     0,     0,     0  ],
            [0,     1,     0,     0,     0,     0,     0  ],
            [0,     0,     0,     0,     0,     0,     1  ],
            [0,     0,     0,     0,     0,     0.5,   0.5],
            [0,     0.5,   0,     0,     0.5,   0,     0  ],
            [0.5,   0.5,   0,     0,     0,     0,     0  ]
        ])
        marchie = MarChie(trans_mat=trans_mat)

        expected_equivalency_cls_mat = np.array([
            [1,     0,     0],
            [0,     1,     1]
        ])
        actual_equivalency_cls_mat = marchie.equivalency_cls_mat.original

        self.assertTrue(
            np.array_equal(expected_equivalency_cls_mat, actual_equivalency_cls_mat)
        )
        self.assertIs(actual_equivalency_cls_mat.dtype, np.dtype(int))

    #endregion


class EndBehaviourTest(unittest.TestCase):

    '''
    ### Tests properties, structure, classification and end behavior of a Markov Chain.
    '''

    #region Ergodic Chains

    # ergodic regular chain 
    def test_example_ergodic_marchie_0(self):

        # create the test chain
        trans_mat = np.array([
            [0.5,    0.25,   0.25],
            [0.5,    0,      0.5 ],
            [0.25,   0.25,   0.5 ]
        ])
        marchie = MarChie(trans_mat=trans_mat)

        # test canonical numbering: should be the same as we have 
        # a single cyclic subclass and no inessential states here
        expected_canonical_mapping = {0: 0, 1: 1, 2: 2}
        expected_canonical_trans_mat = np.array([
            [0.5,    0.25,   0.25],
            [0.5,    0,      0.5 ],
            [0.25,   0.25,   0.5 ]
        ])
        actual_canonical_mapping = marchie.canonical_mapping
        actual_canonical_trans_mat = marchie.trans_mat.canonical
        self.assertDictEqual(expected_canonical_mapping, actual_canonical_mapping)
        self.assertTrue(
            np.array_equal(expected_canonical_trans_mat, actual_canonical_trans_mat)
        )

        # test structure
        self.assertListEqual(marchie.structure.original_inessential_states, [])
        self.assertListEqual(marchie.structure.canonical_inessential_states, [])
        self.assertEqual(marchie.n_equivalency_classes, 1)
        self.assertEqual(marchie.n_cyclic_subclasses, 1)
        self.assertEqual(marchie.structure.equivalency_classes[0].d, 1)
        self.assertListEqual(marchie.structure.cyclic_subclasses[0].original_states, [0, 1, 2])
        self.assertListEqual(marchie.structure.cyclic_subclasses[0].canonical_states, [0, 1, 2])

        # test classification
        self.assertFalse(marchie.properties.reducible)
        self.assertFalse(marchie.properties.polyergodic)
        self.assertTrue(marchie.properties.regular)
        self.assertFalse(marchie.properties.absorbing)
        self.assertTrue(marchie.properties.strong_convergence)
        self.assertIs(marchie.end_behavior.__class__, ErgodicChainEndBehavior)
        
        # test end behavior
        expected_end_trans_mat = np.array([
            [0.4,    0.2,    0.4],
            [0.4,    0.2,    0.4],
            [0.4,    0.2,    0.4]
        ])
        expected_end_state_distr = np.array(
            [0.4,    0.2,    0.4]
        )
        expected_time_percentage = np.array(
            [40,     20,     40]
        )
        actual_end_trans_mat = marchie.end_behavior.end_trans_mat
        actual_end_state_distr = marchie.end_behavior.end_state_distr
        actual_time_percentage = marchie.end_behavior.time_percentage.round(2)
        self.assertTrue(
            np.allclose(expected_end_trans_mat, actual_end_trans_mat)
        )
        self.assertTrue(
            np.allclose(expected_end_state_distr, actual_end_state_distr)
        )
        self.assertTrue(
            np.allclose(expected_time_percentage, actual_time_percentage)
        )
        # here the matrix in a large number of steps is the best test
        # since it's actual empiric end values
        self.assertTrue(
            np.allclose(actual_end_trans_mat, marchie.trans_insteps(n_steps=100000))
        )
        self.assertTrue(
            np.allclose(actual_end_state_distr, marchie.states_insteps(n_steps=100000))
        )

    # ergodic cyclic chain
    def test_example_ergodic_marchie_1(self):

        # create the test chain
        trans_mat = np.array([
            [0,   1,     0,     0,     0  ],
            [0.3, 0,     0.7,   0,     0  ],
            [0,   0.3,   0,     0.7,   0  ],
            [0,   0,     0.3,   0,     0.7],
            [0,   0,     0,     1,     0  ],
        ])
        marchie = MarChie(trans_mat=trans_mat)

        # test canonical numbering
        expected_canonical_mapping = {0: 2, 1: 0, 2: 3, 3: 1, 4: 4}
        expected_canonical_trans_mat = np.array([
            [0,   0,     0.3,   0.7,   0  ],
            [0,   0,     0,     0.3,   0.7],
            [1,   0,     0,     0,     0  ],
            [0.3, 0.7,   0,     0,     0  ],
            [0,   1,     0,     0,     0  ]
        ])
        actual_canonical_mapping = marchie.canonical_mapping
        actual_canonical_trans_mat = marchie.trans_mat.canonical
        self.assertDictEqual(expected_canonical_mapping, actual_canonical_mapping)
        self.assertTrue(
            np.array_equal(expected_canonical_trans_mat, actual_canonical_trans_mat)
        )

        # test structure
        self.assertListEqual(marchie.structure.original_inessential_states, [])
        self.assertListEqual(marchie.structure.canonical_inessential_states, [])
        self.assertEqual(marchie.n_equivalency_classes, 1)
        self.assertEqual(marchie.n_cyclic_subclasses, 2)
        self.assertEqual(marchie.structure.equivalency_classes[0].d, 2)
        self.assertListEqual(marchie.structure.cyclic_subclasses[0].original_states, [1, 3])
        self.assertListEqual(marchie.structure.cyclic_subclasses[0].canonical_states, [0, 1])
        self.assertListEqual(marchie.structure.cyclic_subclasses[1].original_states, [0, 2, 4])
        self.assertListEqual(marchie.structure.cyclic_subclasses[1].canonical_states, [2, 3, 4])

        # test classification
        self.assertFalse(marchie.properties.reducible)
        self.assertFalse(marchie.properties.polyergodic)
        self.assertFalse(marchie.properties.regular)
        self.assertFalse(marchie.properties.absorbing)
        self.assertFalse(marchie.properties.strong_convergence)
        self.assertIs(marchie.end_behavior.__class__, ErgodicChainEndBehavior)
        
        # test end behavior
        expected_end_trans_mat = np.array([
            [0.07758621,   0.42241379,   0.02327586,   0.18103448,   0.29568966],
            [0.07758621,   0.42241379,   0.02327586,   0.18103448,   0.29568966],
            [0.07758621,   0.42241379,   0.02327586,   0.18103448,   0.29568966],
            [0.07758621,   0.42241379,   0.02327586,   0.18103448,   0.29568966],
            [0.07758621,   0.42241379,   0.02327586,   0.18103448,   0.29568966]
        ])
        expected_end_state_distr = np.array(
            [0.07758621,   0.42241379,   0.02327586,   0.18103448,   0.29568966]
        )
        expected_time_percentage = np.array(
            [7.76,   42.24,   2.33,   18.1,   29.57],
        )
        actual_end_trans_mat = marchie.end_behavior.end_trans_mat
        actual_end_state_distr = marchie.end_behavior.end_state_distr
        actual_time_percentage = marchie.end_behavior.time_percentage.round(2)
        self.assertTrue(
            np.allclose(expected_end_trans_mat, actual_end_trans_mat)
        )
        self.assertTrue(
            np.allclose(expected_end_state_distr, actual_end_state_distr)
        )
        self.assertTrue(
            np.allclose(expected_time_percentage, actual_time_percentage)
        )
        # convergence is weak
        self.assertFalse(
            np.allclose(actual_end_trans_mat, marchie.trans_insteps(n_steps=100000))
        )
        self.assertFalse(
            np.allclose(actual_end_state_distr, marchie.states_insteps(n_steps=100000))
        )

    #endregion

    #region Monoergodic Chains

    # monoergodic regular chain 
    def test_example_monoergodic_marchie_0(self):

        # create the test chain
        trans_mat = np.array([
            [0,      0.5,    0,      0.5,    0],
            [0.5,    0,      0,      0,    0.5],
            [0,      0,      0.5,    0,    0.5],
            [0,      0,      1,      0,      0],
            [0,      0,      0,      1,      0]
        ])
        marchie = MarChie(trans_mat=trans_mat)

        # test canonical numbering
        expected_canonical_mapping = {0: 3, 1: 4, 2: 0, 3: 1, 4: 2}
        expected_canonical_trans_mat = np.array([
            [0.5,    0,      0.5,    0,      0],
            [1,      0,      0,      0,      0],
            [0,      1,      0,      0,      0],
            [0,      0.5,    0,      0,    0.5],
            [0,      0,      0.5,    0.5,    0]
        ])
        actual_canonical_mapping = marchie.canonical_mapping
        actual_canonical_trans_mat = marchie.trans_mat.canonical
        self.assertDictEqual(expected_canonical_mapping, actual_canonical_mapping)
        self.assertTrue(
            np.array_equal(expected_canonical_trans_mat, actual_canonical_trans_mat)
        )

        # test structure
        self.assertListEqual(marchie.structure.original_inessential_states, [0, 1])
        self.assertListEqual(marchie.structure.canonical_inessential_states, [3, 4])
        self.assertEqual(marchie.n_equivalency_classes, 1)
        self.assertEqual(marchie.n_cyclic_subclasses, 1)
        self.assertEqual(marchie.structure.equivalency_classes[0].d, 1)
        self.assertListEqual(marchie.structure.cyclic_subclasses[0].original_states, [2, 3, 4])
        self.assertListEqual(marchie.structure.cyclic_subclasses[0].canonical_states, [0, 1, 2])

        # test classification
        self.assertTrue(marchie.properties.reducible)
        self.assertFalse(marchie.properties.polyergodic)
        self.assertTrue(marchie.properties.regular)
        self.assertFalse(marchie.properties.absorbing)
        self.assertTrue(marchie.properties.strong_convergence)
        self.assertIs(marchie.end_behavior.__class__, MonoergodicChainEndBehavior)
        
        # test end behavior
        expected_end_trans_mat = np.array([
            [0.5,    0.25,    0.25,   0,     0],
            [0.5,    0.25,    0.25,   0,     0],
            [0.5,    0.25,    0.25,   0,     0],
            [0.5,    0.25,    0.25,   0,     0],
            [0.5,    0.25,    0.25,   0,     0]
        ])
        expected_end_state_distr = np.array(
            [0.5,    0.25,    0.25,   0,     0]
        )
        expected_time_percentage = np.array(
            [50,     25,      25,     0,     0]
        )
        actual_end_trans_mat = marchie.end_behavior.end_trans_mat
        actual_end_state_distr = marchie.end_behavior.end_state_distr
        actual_time_percentage = marchie.end_behavior.time_percentage.round(2)
        self.assertTrue(
            np.allclose(expected_end_trans_mat, actual_end_trans_mat)
        )
        self.assertTrue(
            np.allclose(expected_end_state_distr, actual_end_state_distr)
        )
        self.assertTrue(
            np.allclose(expected_time_percentage, actual_time_percentage)
        )
        # here the matrix in a large number of steps is the best test
        # since it's actual empiric end values
        self.assertTrue(
            np.allclose(actual_end_trans_mat, marchie.trans_insteps(n_steps=100000))
        )
        self.assertTrue(
            np.allclose(actual_end_state_distr, marchie.states_insteps(n_steps=100000))
        )

    # monoergodic cyclic chain 
    def test_example_monoergodic_marchie_0(self):

        # create the test chain
        trans_mat = np.array([
            [0.5,    0.5,    0,      0,      0,      0  ],
            [0.25,   0.25,   0.25,   0,      0.25,   0  ],
            [0,      0,      0,      0,      1,      0  ],
            [0,      0,      0.5,    0.5,    0,      0  ],
            [0,      0,      1,      0,      0,      0  ],
            [0,      0.25,   0,      0.25,   0.5,    0  ]
        ])
        marchie = MarChie(trans_mat=trans_mat)

        # test canonical numbering
        expected_canonical_mapping = { 0: 2, 1: 3, 2: 0, 3: 4, 4: 1, 5: 5}
        expected_canonical_trans_mat = np.array([
            [0,      1,      0,      0,      0,      0  ],
            [1,      0,      0,      0,      0,      0  ],
            [0,      0,      0.5,    0.5,    0,      0  ],
            [0.25,   0.25,   0.25,   0.25,   0,      0  ],
            [0.5,    0,      0,      0,      0.5,    0  ],
            [0,      0.5,    0,      0.25,   0.25,   0  ]
        ])
        actual_canonical_mapping = marchie.canonical_mapping
        actual_canonical_trans_mat = marchie.trans_mat.canonical
        self.assertDictEqual(expected_canonical_mapping, actual_canonical_mapping)
        self.assertTrue(
            np.array_equal(expected_canonical_trans_mat, actual_canonical_trans_mat)
        )

        # test structure
        self.assertListEqual(marchie.structure.original_inessential_states, [0, 1, 3, 5])
        self.assertListEqual(marchie.structure.canonical_inessential_states, [2, 3, 4, 5])
        self.assertEqual(marchie.n_equivalency_classes, 1)
        self.assertEqual(marchie.n_cyclic_subclasses, 2)
        self.assertEqual(marchie.structure.equivalency_classes[0].d, 2)
        self.assertListEqual(marchie.structure.cyclic_subclasses[0].original_states, [2])
        self.assertListEqual(marchie.structure.cyclic_subclasses[0].canonical_states, [0])
        self.assertListEqual(marchie.structure.cyclic_subclasses[1].original_states, [4])
        self.assertListEqual(marchie.structure.cyclic_subclasses[1].canonical_states, [1])

        # test classification
        self.assertTrue(marchie.properties.reducible)
        self.assertFalse(marchie.properties.polyergodic)
        self.assertFalse(marchie.properties.regular)
        self.assertFalse(marchie.properties.absorbing)
        self.assertFalse(marchie.properties.strong_convergence)
        self.assertIs(marchie.end_behavior.__class__, MonoergodicChainEndBehavior)
        
        # test end behavior
        expected_end_trans_mat = np.array([
            [0.5,    0.5,     0,      0,     0,     0],
            [0.5,    0.5,     0,      0,     0,     0],
            [0.5,    0.5,     0,      0,     0,     0],
            [0.5,    0.5,     0,      0,     0,     0],
            [0.5,    0.5,     0,      0,     0,     0],
            [0.5,    0.5,     0,      0,     0,     0]
        ])
        expected_end_state_distr = np.array(
            [0.5,    0.5,     0,      0,     0,     0]
        )
        expected_time_percentage = np.array(
            [50,     50,      0,      0,     0,     0]
        )
        actual_end_trans_mat = marchie.end_behavior.end_trans_mat
        actual_end_state_distr = marchie.end_behavior.end_state_distr
        actual_time_percentage = marchie.end_behavior.time_percentage.round(2)
        self.assertTrue(
            np.allclose(expected_end_trans_mat, actual_end_trans_mat)
        )
        self.assertTrue(
            np.allclose(expected_end_state_distr, actual_end_state_distr)
        )
        self.assertTrue(
            np.allclose(expected_time_percentage, actual_time_percentage)
        )
        # convergence is weak
        self.assertFalse(
            np.allclose(actual_end_trans_mat, marchie.trans_insteps(n_steps=100000))
        )
        self.assertFalse(
            np.allclose(actual_end_state_distr, marchie.states_insteps(n_steps=100000))
        )

    # monoergodic absorbing chain 
    def test_example_monoergodic_marchie_1(self):

        # create the test chain
        trans_mat = np.array([
            [1,      0,      0,      0,      0   ],
            [0.77,   0,      0.23,   0,      0   ],
            [0,      0.77,   0,      0.23,   0   ],
            [0,      0,      0.77,   0,      0.23],
            [0,      0,      0,      1,      0   ]
        ])
        marchie = MarChie(trans_mat=trans_mat)

        # test canonical numbering
        expected_canonical_mapping = {0: 0, 1: 1, 2: 2, 3: 3, 4: 4}
        expected_canonical_trans_mat = np.array([
            [1,      0,      0,      0,      0   ],
            [0.77,   0,      0.23,   0,      0   ],
            [0,      0.77,   0,      0.23,   0   ],
            [0,      0,      0.77,   0,      0.23],
            [0,      0,      0,      1,      0   ]
        ])
        actual_canonical_mapping = marchie.canonical_mapping
        actual_canonical_trans_mat = marchie.trans_mat.canonical
        self.assertDictEqual(expected_canonical_mapping, actual_canonical_mapping)
        self.assertTrue(
            np.array_equal(expected_canonical_trans_mat, actual_canonical_trans_mat)
        )

        # test structure
        self.assertListEqual(marchie.structure.original_inessential_states, [1, 2, 3, 4])
        self.assertListEqual(marchie.structure.canonical_inessential_states, [1, 2, 3, 4])
        self.assertEqual(marchie.n_equivalency_classes, 1)
        self.assertEqual(marchie.n_cyclic_subclasses, 1)
        self.assertEqual(marchie.structure.equivalency_classes[0].d, 1)
        self.assertListEqual(marchie.structure.cyclic_subclasses[0].original_states, [0])
        self.assertListEqual(marchie.structure.cyclic_subclasses[0].canonical_states, [0])

        # test classification
        self.assertTrue(marchie.properties.reducible)
        self.assertFalse(marchie.properties.polyergodic)
        self.assertTrue(marchie.properties.regular)
        self.assertTrue(marchie.properties.absorbing)
        self.assertTrue(marchie.properties.strong_convergence)
        self.assertIs(marchie.end_behavior.__class__, MonoergodicChainEndBehavior)
        
        # test end behavior
        expected_end_trans_mat = np.array([
            [1,      0,      0,      0,      0],
            [1,      0,      0,      0,      0],
            [1,      0,      0,      0,      0],
            [1,      0,      0,      0,      0],
            [1,      0,      0,      0,      0]
        ])
        expected_end_state_distr = np.array(
            [1,      0,      0,      0,      0]
        )
        expected_time_percentage = np.array(
            [100,    0,      0,      0,      0]
        )
        actual_end_trans_mat = marchie.end_behavior.end_trans_mat
        actual_end_state_distr = marchie.end_behavior.end_state_distr
        actual_time_percentage = marchie.end_behavior.time_percentage.round(2)
        self.assertTrue(
            np.allclose(expected_end_trans_mat, actual_end_trans_mat)
        )
        self.assertTrue(
            np.allclose(expected_end_state_distr, actual_end_state_distr)
        )
        self.assertTrue(
            np.allclose(expected_time_percentage, actual_time_percentage)
        )
        # here the matrix in a large number of steps is the best test
        # since it's actual empiric end values
        self.assertTrue(
            np.allclose(actual_end_trans_mat, marchie.trans_insteps(n_steps=100000))
        )
        self.assertTrue(
            np.allclose(actual_end_state_distr, marchie.states_insteps(n_steps=100000))
        )

    #endregion

    #region Polyergodic Chains

    # polyergodic regular chain 
    def test_example_polyergodic_marchie_0(self):

        # create the test chain
        trans_mat = np.array([            
            [1/3,   0,     0,     0,     0,     2/3,   0,     0,     0  ],   
            [0,     0,     0.5,   0,     0,     0,     0,     0.5,   0  ],   
            [0,     0.5,   0,     0,     0,     0,     0,     0.5,   0  ],
            [0,     0,     0,     1,     0,     0,     0,     0,     0  ],   
            [0,     0,     1/3,   1/3,   0,     0,     0,     1/3,   0  ],   
            [1,     0,     0,     0,     0,     0,     0,     0,     0  ],
            [1/3,   0,     0,     0,     1/3,   0,     0,     0,     1/3],   
            [0,     0.5,   0.5,   0,     0,     0,     0,     0,     0  ],   
            [1/3,   0,     0,     0,     0,     1/3,   1/3,   0,     0  ]
        ])
        marchie = MarChie(trans_mat=trans_mat)

        # test canonical numbering
        expected_canonical_mapping = {0: 0, 1: 2, 2: 3, 3: 5, 4: 6, 5: 1, 6: 7, 7: 4, 8: 8}
        expected_canonical_trans_mat = np.array([
            [1/3,   2/3,   0,     0,     0,     0,     0,     0,     0  ],   
            [1,     0,     0,     0,     0,     0,     0,     0,     0  ],   
            [0,     0,     0,     0.5,   0.5,   0,     0,     0,     0  ],
            [0,     0,     0.5,   0,     0.5,   0,     0,     0,     0  ],   
            [0,     0,     0.5,   0.5,   0,     0,     0,     0,     0  ],   
            [0,     0,     0,     0,     0,     1,     0,     0,     0  ],
            [0,     0,     0,     1/3,   1/3,   1/3,   0,     0,     0  ],   
            [1/3,   0,     0,     0,     0,     0,     1/3,   0,     1/3],   
            [1/3,   1/3,   0,     0,     0,     0,     0,     1/3,   0  ]
        ])
        actual_canonical_mapping = marchie.canonical_mapping
        actual_canonical_trans_mat = marchie.trans_mat.canonical
        self.assertDictEqual(expected_canonical_mapping, actual_canonical_mapping)
        self.assertTrue(
            np.array_equal(expected_canonical_trans_mat, actual_canonical_trans_mat)
        )

        # test structure
        self.assertListEqual(marchie.structure.original_inessential_states, [4, 6, 8])
        self.assertListEqual(marchie.structure.canonical_inessential_states, [6, 7, 8])
        self.assertEqual(marchie.n_equivalency_classes, 3)
        self.assertEqual(marchie.n_cyclic_subclasses, 3)
        self.assertEqual(marchie.structure.equivalency_classes[0].d, 1)
        self.assertEqual(marchie.structure.equivalency_classes[1].d, 1)
        self.assertEqual(marchie.structure.equivalency_classes[2].d, 1)
        self.assertListEqual(marchie.structure.cyclic_subclasses[0].original_states, [0, 5])
        self.assertListEqual(marchie.structure.cyclic_subclasses[0].canonical_states, [0, 1])
        self.assertListEqual(marchie.structure.cyclic_subclasses[1].original_states, [1, 2, 7])
        self.assertListEqual(marchie.structure.cyclic_subclasses[1].canonical_states, [2, 3, 4])
        self.assertListEqual(marchie.structure.cyclic_subclasses[2].original_states, [3])
        self.assertListEqual(marchie.structure.cyclic_subclasses[2].canonical_states, [5])

        # test classification
        self.assertTrue(marchie.properties.reducible)
        self.assertTrue(marchie.properties.polyergodic)
        self.assertTrue(marchie.properties.regular)
        self.assertFalse(marchie.properties.absorbing)
        self.assertTrue(marchie.properties.strong_convergence)
        self.assertIs(marchie.end_behavior.__class__, PolyergodicChainEndBehavior)
        
        # test end behavior
        expected_end_trans_mat = np.array([
            [0.6,   0.4,   0,     0,     0,     0,     0,     0,     0  ],   
            [0.6,   0.4,   0,     0,     0,     0,     0,     0,     0  ],   
            [0,     0,     1/3,   1/3,   1/3,   0,     0,     0,     0  ],
            [0,     0,     1/3,   1/3,   1/3,   0,     0,     0,     0  ],   
            [0,     0,     1/3,   1/3,   1/3,   0,     0,     0,     0  ],   
            [0,     0,     0,     0,     0,     1,     0,     0,     0  ],
            [0,     0,     2/9,   2/9,   2/9,   1/3,   0,     0,     0  ],   
            [3/8,   0.25,  1/12,  1/12,  1/12,  0.125, 0,     0,     0  ],   
            [21/40, 14/40, 1/36,  1/36,  1/36,  1/24,  0,     0,     0  ]
        ])
        actual_end_trans_mat = marchie.end_behavior.end_trans_mat
        actual_end_state_distr = marchie.end_behavior.end_state_distr
        self.assertTrue(
            np.allclose(expected_end_trans_mat, actual_end_trans_mat)
        )
        self.assertIsNone(marchie.end_behavior.time_percentage) # not supposed for polyergodic chains
        # here the matrix in a large number of steps is the best test
        # since it's actual empiric end values
        self.assertTrue(
            np.allclose(actual_end_trans_mat, marchie.trans_insteps(n_steps=100000))
        )
        self.assertTrue(
            np.allclose(actual_end_state_distr, marchie.states_insteps(n_steps=100000))
        )

    # polyergodic cyclic chain 
    def test_example_polyergodic_marchie_1(self):

        # create the test chain
        trans_mat = np.array([            
            [0,     0.3,   0.7,   0,     0,     0,     0,     0,     0,     0,     0],
            [0,     0,     0,     1,     0,     0,     0,     0,     0,     0,     0],
            [1,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0],
            [0,     0,     0,     0,     0.2,   0,     0,     0.8,   0,     0,     0],
            [0,     0,     0,     0,     0,     1,     0,     0,     0,     0,     0],
            [0,     0,     0,     0,     0,     0,     1,     0,     0,     0,     0],
            [0,     0,     0,     0,     1,     0,     0,     0,     0,     0,     0],
            [0,     0,     0,     0,     0,     0,     0,     0,     1,     0,     0],
            [0,     0,     0,     0,     0,     0,     0,     0,     0,     1,     0],
            [0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     1],
            [0,     0,     0,     0,     0,     0,     0,     1,     0,     0,     0]
        ])
        marchie = MarChie(trans_mat=trans_mat)

        # test canonical numbering
        expected_canonical_mapping = {0: 7, 1: 8, 2: 9, 3: 10, 4: 0, 5: 1, 6: 2, 7: 3, 8: 4, 9: 5, 10: 6}
        expected_canonical_trans_mat = np.array([
            [0,     1,     0,     0,     0,     0,     0,     0,     0,     0,     0],
            [0,     0,     1,     0,     0,     0,     0,     0,     0,     0,     0],
            [1,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0],
            [0,     0,     0,     0,     1,     0,     0,     0,     0,     0,     0],
            [0,     0,     0,     0,     0,     1,     0,     0,     0,     0,     0],
            [0,     0,     0,     0,     0,     0,     1,     0,     0,     0,     0],
            [0,     0,     0,     1,     0,     0,     0,     0,     0,     0,     0],
            [0,     0,     0,     0,     0,     0,     0,     0,     0.3,   0.7,   0],
            [0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     1],
            [0,     0,     0,     0,     0,     0,     0,     1,     0,     0,     0],
            [0.2,   0,     0,     0.8,   0,     0,     0,     0,     0,     0,     0]
        ])
        actual_canonical_mapping = marchie.canonical_mapping
        actual_canonical_trans_mat = marchie.trans_mat.canonical
        self.assertDictEqual(expected_canonical_mapping, actual_canonical_mapping)
        self.assertTrue(
            np.array_equal(expected_canonical_trans_mat, actual_canonical_trans_mat)
        )

        # test structure
        self.assertListEqual(marchie.structure.original_inessential_states, [0, 1, 2, 3])
        self.assertListEqual(marchie.structure.canonical_inessential_states, [7, 8, 9, 10])
        self.assertEqual(marchie.n_equivalency_classes, 2)
        self.assertEqual(marchie.n_cyclic_subclasses, 7)
        self.assertEqual(marchie.structure.equivalency_classes[0].d, 3)
        self.assertEqual(marchie.structure.equivalency_classes[1].d, 4)
        self.assertListEqual(marchie.structure.cyclic_subclasses[0].original_states, [4])
        self.assertListEqual(marchie.structure.cyclic_subclasses[0].canonical_states, [0])
        self.assertListEqual(marchie.structure.cyclic_subclasses[1].original_states, [5])
        self.assertListEqual(marchie.structure.cyclic_subclasses[1].canonical_states, [1])
        self.assertListEqual(marchie.structure.cyclic_subclasses[2].original_states, [6])
        self.assertListEqual(marchie.structure.cyclic_subclasses[2].canonical_states, [2])
        self.assertListEqual(marchie.structure.cyclic_subclasses[3].original_states, [7])
        self.assertListEqual(marchie.structure.cyclic_subclasses[3].canonical_states, [3])
        self.assertListEqual(marchie.structure.cyclic_subclasses[4].original_states, [8])
        self.assertListEqual(marchie.structure.cyclic_subclasses[4].canonical_states, [4])
        self.assertListEqual(marchie.structure.cyclic_subclasses[5].original_states, [9])
        self.assertListEqual(marchie.structure.cyclic_subclasses[5].canonical_states, [5])
        self.assertListEqual(marchie.structure.cyclic_subclasses[6].original_states, [10])
        self.assertListEqual(marchie.structure.cyclic_subclasses[6].canonical_states, [6])

        # test classification
        self.assertTrue(marchie.properties.reducible)
        self.assertTrue(marchie.properties.polyergodic)
        self.assertFalse(marchie.properties.regular)
        self.assertFalse(marchie.properties.absorbing)
        self.assertFalse(marchie.properties.strong_convergence)
        self.assertIs(marchie.end_behavior.__class__, PolyergodicChainEndBehavior)
        
        # test end behavior
        expected_end_trans_mat = np.array([
            [1/3,   1/3,   1/3,   0,     0,     0,     0,     0,     0,     0,     0],
            [1/3,   1/3,   1/3,   0,     0,     0,     0,     0,     0,     0,     0],
            [1/3,   1/3,   1/3,   0,     0,     0,     0,     0,     0,     0,     0],
            [0,     0,     0,     0.25,  0.25,  0.25,  0.25,  0,     0,     0,     0],
            [0,     0,     0,     0.25,  0.25,  0.25,  0.25,  0,     0,     0,     0],
            [0,     0,     0,     0.25,  0.25,  0.25,  0.25,  0,     0,     0,     0],
            [0,     0,     0,     0.25,  0.25,  0.25,  0.25,  0,     0,     0,     0],
            [1/15,  1/15,  1/15,  0.2,   0.2,   0.2,   0.2,   0,     0,     0,     0],
            [1/15,  1/15,  1/15,  0.2,   0.2,   0.2,   0.2,   0,     0,     0,     0],
            [1/15,  1/15,  1/15,  0.2,   0.2,   0.2,   0.2,   0,     0,     0,     0],
            [1/15,  1/15,  1/15,  0.2,   0.2,   0.2,   0.2,   0,     0,     0,     0]
        ])
        actual_end_trans_mat = marchie.end_behavior.end_trans_mat
        actual_end_state_distr = marchie.end_behavior.end_state_distr
        self.assertTrue(
            np.allclose(expected_end_trans_mat, actual_end_trans_mat)
        )
        self.assertIsNone(marchie.end_behavior.time_percentage) # not supposed for polyergodic chains
        # convergence is weak
        self.assertFalse(
            np.allclose(actual_end_trans_mat, marchie.trans_insteps(n_steps=100000))
        )
        self.assertFalse(
            np.allclose(actual_end_state_distr, marchie.states_insteps(n_steps=100000))
        )

    # polyergodic absorbing chain 
    def test_example_polyergodic_absorbing_marchie_0(self):

         # create the test chain
        trans_mat = np.array([            
            [1,    0,     0,     0,     0  ],
            [0.3,  0,     0.7,   0,     0  ],
            [0,    0.3,   0,     0.7,   0  ],
            [0,    0,     0.3,   0,     0.7],
            [0,    0,     0,     0,     1  ]
        ])
        marchie = MarChie(trans_mat=trans_mat)

        # test canonical numbering
        expected_canonical_mapping = {0: 0, 1: 2, 2: 3, 3: 4, 4: 1}
        expected_canonical_trans_mat = np.array([
            [1,    0,     0,     0,     0  ],
            [0,    1,     0,     0,     0  ],
            [0.3,  0,     0,     0.7,   0  ],
            [0,    0,     0.3,   0,     0.7],
            [0,    0.7,   0,     0.3,   0  ]
        ])
        actual_canonical_mapping = marchie.canonical_mapping
        actual_canonical_trans_mat = marchie.trans_mat.canonical
        self.assertDictEqual(expected_canonical_mapping, actual_canonical_mapping)
        self.assertTrue(
            np.array_equal(expected_canonical_trans_mat, actual_canonical_trans_mat)
        )

        # test structure
        self.assertListEqual(marchie.structure.original_inessential_states, [1, 2, 3])
        self.assertListEqual(marchie.structure.canonical_inessential_states, [2, 3, 4])
        self.assertEqual(marchie.n_equivalency_classes, 2)
        self.assertEqual(marchie.n_cyclic_subclasses, 2)
        self.assertEqual(marchie.structure.equivalency_classes[0].d, 1)
        self.assertEqual(marchie.structure.equivalency_classes[1].d, 1)
        self.assertListEqual(marchie.structure.cyclic_subclasses[0].original_states, [0])
        self.assertListEqual(marchie.structure.cyclic_subclasses[0].canonical_states, [0])
        self.assertListEqual(marchie.structure.cyclic_subclasses[1].original_states, [4])
        self.assertListEqual(marchie.structure.cyclic_subclasses[1].canonical_states, [1])

        # test classification
        self.assertTrue(marchie.properties.reducible)
        self.assertTrue(marchie.properties.polyergodic)
        self.assertTrue(marchie.properties.regular)
        self.assertTrue(marchie.properties.absorbing)
        self.assertTrue(marchie.properties.strong_convergence)
        self.assertIs(marchie.end_behavior.__class__, PolyergodicAbsorbingChainEndBehavior)
        
        # test end behavior
        expected_end_trans_mat = np.array([
            [1,           0,            0,     0,     0],
            [0,           1,            0,     0,     0],
            [0.40862069,  0.59137931,   0,     0,     0],
            [0.15517241,  0.84482759,   0,     0,     0],
            [0.04655172,  0.95344828,   0,     0,     0]
        ])
        actual_end_trans_mat = marchie.end_behavior.end_trans_mat
        actual_end_state_distr = marchie.end_behavior.end_state_distr
        self.assertTrue(
            np.allclose(expected_end_trans_mat, actual_end_trans_mat)
        )
        self.assertIsNone(marchie.end_behavior.time_percentage) # not supposed for polyergodic chains
        # here the matrix in a large number of steps is the best test
        # since it's actual empiric end values
        self.assertTrue(
            np.allclose(actual_end_trans_mat, marchie.trans_insteps(n_steps=100000))
        )
        self.assertTrue(
            np.allclose(actual_end_state_distr, marchie.states_insteps(n_steps=100000))
        )

    # polyergodic absorbing chain 
    def test_example_polyergodic_absorbing_marchie_1(self):

        # create the test chain
        trans_mat = np.array([            
            [1/3,  0,     0,     0,     1/3,   1/3],
            [1/3,  1/3,   0,     0,     1/3,   0  ],
            [0,    1/3,   1/3,   0,     1/3,   0  ], 
            [0,    0,     1/3,   1/3,   1/3,   0  ],
            [0,    0,     0,     0,     1,     0  ],
            [0,    0,     0,     0,     0,     1  ]
        ])
        marchie = MarChie(trans_mat=trans_mat)

        # test canonical numbering
        expected_canonical_mapping = {4: 0, 5: 1, 0: 2, 1: 3, 2: 4, 3: 5}
        expected_canonical_trans_mat = np.array([
            [1,      0,     0,     0,     0,     0  ],
            [0,      1,     0,     0,     0,     0  ],
            [1/3,    1/3,   1/3,   0,     0,     0  ], 
            [1/3,    0,     1/3,   1/3,   0,     0  ],
            [1/3,    0,     0,     1/3,   1/3,   0  ],
            [1/3,    0,     0,     0,     1/3,   1/3]
        ])
        actual_canonical_mapping = marchie.canonical_mapping
        actual_canonical_trans_mat = marchie.trans_mat.canonical
        self.assertDictEqual(expected_canonical_mapping, actual_canonical_mapping)
        self.assertTrue(
            np.array_equal(expected_canonical_trans_mat, actual_canonical_trans_mat)
        )

        # test structure
        self.assertListEqual(marchie.structure.original_inessential_states, [0, 1, 2, 3])
        self.assertListEqual(marchie.structure.canonical_inessential_states, [2, 3, 4, 5])
        self.assertEqual(marchie.n_equivalency_classes, 2)
        self.assertEqual(marchie.n_cyclic_subclasses, 2)
        self.assertEqual(marchie.structure.equivalency_classes[0].d, 1)
        self.assertEqual(marchie.structure.equivalency_classes[1].d, 1)
        self.assertListEqual(marchie.structure.cyclic_subclasses[0].original_states, [4])
        self.assertListEqual(marchie.structure.cyclic_subclasses[0].canonical_states, [0])
        self.assertListEqual(marchie.structure.cyclic_subclasses[1].original_states, [5])
        self.assertListEqual(marchie.structure.cyclic_subclasses[1].canonical_states, [1])

        # test classification
        self.assertTrue(marchie.properties.reducible)
        self.assertTrue(marchie.properties.polyergodic)
        self.assertTrue(marchie.properties.regular)
        self.assertTrue(marchie.properties.absorbing)
        self.assertTrue(marchie.properties.strong_convergence)
        self.assertIs(marchie.end_behavior.__class__, PolyergodicAbsorbingChainEndBehavior)
        
        # test end behavior
        expected_end_trans_mat = np.array([
            [1,      0,     0,     0,     0,     0  ],
            [0,      1,     0,     0,     0,     0  ],
            [0.5,    0.5,   0,     0,     0,     0  ], 
            [0.75,   0.25,  0,     0,     0,     0  ],
            [0.875,  0.125, 0,     0,     0,     0  ],
            [0.9375, 0.0625,0,     0,     0,     0  ]
        ])
        actual_end_trans_mat = marchie.end_behavior.end_trans_mat
        actual_end_state_distr = marchie.end_behavior.end_state_distr
        self.assertTrue(
            np.allclose(expected_end_trans_mat, actual_end_trans_mat)
        )
        self.assertIsNone(marchie.end_behavior.time_percentage) # not supposed for polyergodic chains
        # here the matrix in a large number of steps is the best test
        # since it's actual empiric end values
        self.assertTrue(
            np.allclose(actual_end_trans_mat, marchie.trans_insteps(n_steps=100000))
        )
        self.assertTrue(
            np.allclose(actual_end_state_distr, marchie.states_insteps(n_steps=100000))
        )

    #endregion

#endregion


#region Executive

if __name__ == '__main__':
    unittest.main()

#endregion


import setuptools

desc = '''
    MarChie: a Compact Open Source Tool for Analyzing Discrete Markov Chains.
    '''
with open('./README.md') as readme: long_description = readme.read()

setuptools.setup(
    name='marchie',
    version='0.1',
    author='Max Schmaltz',
    author_email='schmaltzmax@gmail.com',
    description=desc,
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='https://github.com/maxschmaltz/MarChie',
    # packages=[
    #     'marchie'
    # ],
    packages = setuptools.find_packages(),
    install_requires=[
        'numpy',
        'graphviz'
    ],
    license='Apache Software License',
    keywords='markov chain, mathematics',
    classifiers=[
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: Apache Software License',
        'Topic :: Scientific/Engineering :: Mathematics',
        'Intended Audience :: Education',
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3 :: Only",
    ]
)